from django.db import models


def clone_field(field):
    """
    Incomplete backport of field.clone (coming in Django 1.7 and South 2)
    """
    if hasattr(field, 'clone'):
        return field.clone()

    cls = type(field)
    args = tuple()
    kwargs = {}

    if cls == models.ForeignKey:
        args = (field.rel.to,) + args

    kwargs['null'] = field.null
    kwargs['blank'] = field.blank
    if getattr(field, 'max_length', None):
        kwargs['max_length'] = field.max_length

    return cls(*args, **kwargs)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models


class Migration(DataMigration):

    def forwards(self, orm):
        if db.backend_name != 'postgres':
            print "Warning: Non-postgres database detected; convenience view will not be created."
            return
        db.execute('''
CREATE OR REPLACE VIEW wq_identifier_joined AS
SELECT
  ct.app_label, ct.model, i.object_id, i.name AS identifier,
  i.is_primary, a.name authority,
  REPLACE(a.object_url, '%%s', i.slug) AS url
FROM wq_identifier i
JOIN django_content_type ct ON i.content_type_id = ct.id
LEFT OUTER JOIN wq_identifiertype a ON i.authority_id = a.id;''')

    def backwards(self, orm):
        if db.backend_name != 'postgres':
            return
        db.execute("DROP VIEW wq_identifier_joined;")

    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'identify.authority': {
            'Meta': {'object_name': 'Authority', 'db_table': "'wq_identifiertype'"},
            'homepage': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'object_url': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        'identify.identifier': {
            'Meta': {'object_name': 'Identifier', 'db_table': "'wq_identifier'"},
            'authority': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['identify.Authority']", 'null': 'True', 'blank': 'True'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_primary': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'db_index': 'True'})
        }
    }

    complete_apps = ['identify']



from django.db.models import *
from .base.models import *
from .annotate.models import *
from .identify.models import *
from .locate.models import *
from .relate.models import *


class IdentifiedRelatedModelManager(
        IdentifiedModelManager, RelatedModelManager):
    pass


class IdentifiedRelatedModel(IdentifiedModel, RelatedModel):
    objects = IdentifiedRelatedModelManager()

    class Meta:
        abstract = True



from rest_framework.renderers import BaseRenderer, JSONPRenderer, JSONRenderer
from django.conf import settings
from wq.db.rest.settings import SRID as DEFAULT_SRID


class JSONRenderer(JSONRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        if renderer_context and 'request' in renderer_context:
            if not renderer_context['request'].is_ajax():
                renderer_context['indent'] = 4
        return super(JSONRenderer, self).render(
            data, accepted_media_type, renderer_context
        )


class AMDRenderer(JSONPRenderer):
    media_type = 'application/javascript'
    format = 'js'
    default_callback = 'define'


class BinaryRenderer(BaseRenderer):
    def render(self, data, media_type=None, render_context=None):
        return data


def binary_renderer(mimetype, extension=None):
    class Renderer(BinaryRenderer):
        media_type = mimetype
        format = extension
    return Renderer


class GeoJSONRenderer(JSONRenderer):
    media_type = 'application/vnd.geo+json'
    format = 'geojson'

    def render(self, data, *args, **kwargs):
        if isinstance(data, list):
            features, simple = self.render_features(data)
            data = {
                'type': 'FeatureCollection',
                'features': features
            }
        elif "list" in data and isinstance(data['list'], list):
            features, simple = self.render_features(data['list'])
            data['type'] = 'FeatureCollection'
            data['features'] = features
            del data['list']

        else:
            data, simple = self.render_feature(data)

        if not simple and getattr(settings, 'SRID', None) != DEFAULT_SRID:
            data['crs'] = {
                'type': 'name',
                'properties': {
                    'name': 'urn:ogc:def:crs:EPSG::%s' % settings.SRID
                }
            }
        return super(GeoJSONRenderer, self).render(data, *args, **kwargs)

    def render_feature(self, obj):
        feature = {
            'type': 'Feature',
            'properties': obj
        }
        simple = False
        if 'id' in obj:
            feature['id'] = obj['id']
            del obj['id']

        if 'latitude' in obj and 'longitude' in obj:
            feature['geometry'] = {
                'type': 'Point',
                'coordinates': [obj['longitude'], obj['latitude']]
            }
            del obj['latitude']
            del obj['longitude']
            simple = True

        elif 'geometry' in obj:
            feature['geometry'] = obj['geometry']
            del obj['geometry']

        elif 'locations' in obj:
            feature['geometry'] = obj['locations']
            del obj['locations']

        return feature, simple

    def render_features(self, objs):
        features = []
        has_simple = False
        for obj in objs:
            feature, simple = self.render_feature(obj)
            if simple:
                has_simple = True
                if feature['geometry']['coordinates'][0] is not None:
                    features.append(feature)
            else:
                features.append(feature)
        return features, has_simple



from django.utils.importlib import import_module
from django.utils.module_loading import module_has_submodule
from django.conf.urls import patterns, include, url
from django.core.paginator import Paginator

from django.contrib.auth.models import AnonymousUser

from django.conf import settings
from rest_framework.routers import DefaultRouter, Route
from rest_framework.urlpatterns import format_suffix_patterns
from rest_framework.settings import api_settings
from rest_framework.response import Response

from .models import ContentType, get_ct
from .permissions import has_perm
from .views import SimpleViewSet, ModelViewSet
from copy import copy


class Router(DefaultRouter):
    _models = set()
    _serializers = {}
    _querysets = {}
    _filters = {}
    _viewsets = {}
    _extra_pages = {}
    _config = {}
    _page_names = {}
    _page_models = {}

    include_root_view = False
    include_config_view = True
    include_multi_view = True

    def __init__(self, trailing_slash=False):
        # Add trailing slash for HTML list views
        self.routes.append(Route(
            url=r'^{prefix}/$',
            mapping={
                'get': 'list',
                'post': 'create'
            },
            name='{basename}-list',
            initkwargs={'suffix': 'List'}
        ))
        super(Router, self).__init__(trailing_slash=trailing_slash)

    def register_model(self, model, viewset=None, serializer=None,
                       queryset=None, filter=None, **kwargs):
        if isinstance(model, basestring) and '.' in model:
            from django.db.models import get_model
            model = get_model(*model.split('.'))
        self._models.add(model)
        ct = get_ct(model)

        if viewset:
            self.register_viewset(model, viewset)
        if serializer:
            self.register_serializer(model, serializer)
        if queryset:
            self.register_queryset(model, queryset)
        if filter:
            self.register_filter(model, filter)

        if 'name' not in kwargs:
            kwargs['name'] = ct.identifier
        if 'url' not in kwargs:
            url = unicode(model._meta.verbose_name_plural)
            kwargs['url'] = url.replace(' ', '')

        self.register_config(model, kwargs)
        self._page_names[model] = kwargs['name']
        self._page_models[kwargs['name']] = model

    def register_viewset(self, model, viewset):
        self._viewsets[model] = viewset

    def register_serializer(self, model, serializer):
        self._serializers[model] = serializer

    def register_queryset(self, model, queryset):
        self._querysets[model] = queryset

    def register_filter(self, model, queryset):
        self._filters[model] = queryset

    def register_config(self, model, config):
        self._config[model] = config
        self._base_config = None

    def update_config(self, model, **kwargs):
        if model not in self._config:
            raise RuntimeError("%s must be registered first" % model)
        self._config[model].update(kwargs)
        self._base_config = None

    def get_serializer_for_model(self, model_class, serializer_depth=None):
        if model_class in self._serializers:
            serializer = self._serializers[model_class]
        else:
            # Make sure we're not dealing with a proxy
            real_model = get_ct(model_class, True).model_class()
            if real_model in self._serializers:
                serializer = self._serializers[real_model]
            else:
                serializer = api_settings.DEFAULT_MODEL_SERIALIZER_CLASS

        class Serializer(serializer):
            class Meta(serializer.Meta):
                depth = serializer_depth
                model = model_class
        return Serializer

    def serialize(self, obj, many=False, depth=None):
        if many:
            # assume obj is a queryset
            model = obj.model
            if depth is None:
                depth = 0
        else:
            model = obj
            if depth is None:
                depth = 1
        serializer = self.get_serializer_for_model(model, depth)
        return serializer(obj, many=many, context={'router': self}).data

    def get_paginate_by_for_model(self, model_class):
        config = self.get_model_config(model_class) or {}
        paginate_by = config.get('per_page', None)
        if paginate_by:
            return paginate_by
        return api_settings.PAGINATE_BY

    def paginate(self, model, page_num, request=None):
        obj_serializer = self.get_serializer_for_model(model)
        paginate_by = self.get_paginate_by_for_model(model)
        viewset = self.get_viewset_for_model(model)
        qs = self.get_queryset_for_model(model, request)
        req = copy(request)
        req.GET = {}
        qs = viewset(
            action="list",
            request=req,
            kwargs={}
        ).filter_queryset(qs)

        paginator = Paginator(qs, paginate_by)
        page = paginator.page(page_num)

        class Serializer(api_settings.DEFAULT_PAGINATION_SERIALIZER_CLASS):
            class Meta:
                object_serializer_class = obj_serializer
        return Serializer(
            instance=page,
            context={'router': self, 'request': request}
        ).data

    def get_queryset_for_model(self, model, request=None):
        if model in self._querysets:
            qs = self._querysets[model]
        else:
            qs = model.objects.all()
        if request and model in self._filters:
            qs = self._filters[model](qs, request)
        return qs

    def get_lookup_for_model(self, model_class):
        if get_ct(model_class).is_identified:
            return 'primary_identifiers__slug'
        else:
            config = self.get_model_config(model_class) or {}
            return config.get('lookup', 'pk')

    def get_viewset_for_model(self, model_class):
        if model_class in self._viewsets:
            viewset = self._viewsets[model_class]
        else:
            # Make sure we're not dealing with a proxy
            real_model = get_ct(model_class, True).model_class()
            if real_model in self._viewsets:
                viewset = self._viewsets[real_model]
            else:
                viewset = ModelViewSet
        lookup = self.get_lookup_for_model(model_class)

        class ViewSet(viewset):
            model = model_class
            router = self
            lookup_field = lookup

        return ViewSet

    _base_config = None

    @property
    def base_config(self):
        if self._base_config:
            return self._base_config

        user = AnonymousUser()
        pages = {}
        for page in self._extra_pages:
            conf, view = self.get_page(page)
            pages[page] = conf
        for model in self._models:
            ct = get_ct(model)
            if not has_perm(user, ct, 'view'):
                continue
            info = self._config[model].copy()
            info['list'] = True
            for perm in ('add', 'change', 'delete'):
                if has_perm(user, ct, perm):
                    info['can_' + perm] = True

            if 'parents' not in info:
                parents = {}
                as_dict = False
                for pct, fields in ct.get_foreign_keys().items():
                    if pct.is_registered():
                        if has_perm(user, pct, 'view'):
                            if len(fields) > 1 or fields[0] != pct.identifier:
                                as_dict = True
                            parents[pct.identifier] = fields
                if as_dict:
                    info['parents'] = parents
                else:
                    info['parents'] = parents.keys()

            if 'children' not in info:
                info['children'] = []
                for cct in ct.get_children():
                    if cct.is_registered():
                        if has_perm(user, cct, 'view'):
                            info['children'].append(cct.identifier)

            for name in ('annotated', 'identified', 'located', 'related'):
                if getattr(ct, 'is_' + name):
                    info[name] = True

            if ct.is_located or ct.has_geo_fields:
                info['map'] = True

            for field in model._meta.fields:
                if field.choices:
                    info.setdefault('choices', {})
                    info['choices'][field.name] = [{
                        'value': val,
                        'label': unicode(label)
                    } for val, label in field.choices]

            for name in ('annotationtype', 'annotation'):
                if ct.identifier != name and getattr(ct, 'is_' + name):
                    pages[name] = {'alias': ct.identifier}

            pages[info['name']] = info

        self._base_config = {'pages': pages}
        return self._base_config

    def get_config(self, user=None):
        if user is None or not user.is_authenticated():
            return self.base_config

        # Add user-specific permissions to configuration
        config = {
            key: val.copy()
            for key, val in self.base_config.items()
        }
        for page, info in config['pages'].items():
            if not info.get('list', False):
                continue
            ct = get_ct(self._page_models[page])
            info = info.copy()
            for perm in ('add', 'change', 'delete'):
                if has_perm(user, ct, perm):
                    info['can_' + perm] = True
            config['pages'][page] = info

        return config

    def add_page(self, name, config, view=None):
        if view is None:
            class PageView(SimpleViewSet):
                template_name = name + '.html'

                def list(self, request, *args, **kwargs):
                    return Response(config)
            view = PageView
        if 'name' not in config:
            config['name'] = name
        if 'url' not in config:
            config['url'] = name
        self._extra_pages[name] = config, view
        self._base_config = None

    def get_page(self, page):
        return self._extra_pages[page]

    def get_page_config(self, name, user=None):
        config = self.get_config(user)
        return config['pages'].get(name, None)

    def get_model_config(self, model, user=None):
        # First, check models registered with API
        if model in self._page_names:
            config = self.get_config(user)
            return config['pages'][self._page_names[model]]

        # Then check config cache directly (in case model was configured but
        # not fully registered as part of API)
        if model in self._config:
            return self._config[model]
        return None

    def model_is_registered(self, model):
        return model in self._models

    def get_config_view(self):
        class ConfigView(SimpleViewSet):
            def list(this, request, *args, **kwargs):
                return Response(self.get_config(request.user))
        return ConfigView

    def get_multi_view(self):
        class MultipleListView(SimpleViewSet):
            def list(this, request, *args, **kwargs):
                urls = request.GET.get('lists', '').split(',')
                return self.get_multi(request, urls)
        return MultipleListView

    def get_multi(self, request, urls):
        conf = self.get_config(request.user)
        conf_by_url = {
            conf['url']: (page, conf)
            for page, conf
            in self.get_config(request.user)['pages'].items()
        }
        result = {}
        for url in urls:
            if url not in conf_by_url:
                continue
            page, conf = conf_by_url[url]
            model = self._page_models[page]
            result[url] = self.paginate(model, 1, request)
        return Response(result)

    def get_urls(self):
        # Register viewsets with DefaultRouter just before returning urls

        # The root viewset (with a url of "") should be registered last
        root = {}

        def register(config, viewset):
            if config['url'] == "":
                root['view'] = viewset
                root['name'] = config['name']
            else:
                self.register(config['url'], viewset, config['name'])

        # List (model) views
        for model in self._models:
            viewset = self.get_viewset_for_model(model)
            config = self._config[model]
            register(config, viewset)

        # Extra/custom viewsets
        for name in self._extra_pages:
            register(*self._extra_pages[name])

        if self.include_config_view:
            # /config.js
            self.register('config', self.get_config_view(), 'config')

        if self.include_multi_view:
            # /multi.json
            self.register('multi', self.get_multi_view(), 'multi')

        urls = super(Router, self).get_urls()

        if root:
            # / - Skip registration and directly generate custom URLs
            urls.extend(self.get_root_urls(root['view'], root['name']))

        return urls

    def get_root_urls(self, viewset, name):
        lookup = self.get_lookup_regex(viewset)
        routes = self.get_routes(viewset)
        urls = []
        for route in routes:
            pattern = route.url.format(
                prefix='',
                lookup=lookup,
                trailing_slash=self.trailing_slash,
            )

            # Remove leading slash from detail view URLs
            pattern = pattern.replace('/', '', 1)

            mapping = self.get_method_map(viewset, route.mapping)
            if not mapping:
                continue
            view = viewset.as_view(mapping, **route.initkwargs)
            urls.append(url(pattern, view, name=name))

        return format_suffix_patterns(urls)

    def get_routes(self, viewset):
        routes = super(Router, self).get_routes(viewset)
        model = getattr(viewset, "model", None)
        if not model:
            return routes

        # Custom routes

        ct = get_ct(model)
        for pct in ct.get_all_parents():
            if not pct.is_registered():
                continue
            if pct.urlbase == '':
                purlbase = ''
            else:
                purlbase = pct.urlbase + '/'

            purl = (
                '^' + purlbase + r'(?P<' + pct.identifier
                + '>[^\/\?]+)/{prefix}{trailing_slash}$'
            )
            routes.append(Route(
                url=purl,
                mapping={'get': 'list'},
                name="{basename}-for-%s" % pct.identifier,
                initkwargs={'suffix': 'List'},
            ))

        for cct in ct.get_all_children():
            if not cct.is_registered():
                continue
            cbase = cct.urlbase
            routes.append(Route(
                url='^%s-by-{prefix}' % cbase,
                mapping={'get': 'list'},
                name="%s-by-%s" % (cct.identifier, ct.identifier),
                initkwargs={'target': cbase, 'suffix': 'List'},
            ))

        return routes

    @property
    def version(self):
        if not hasattr(self, '_version'):
            vtxt = getattr(settings, 'VERSION_TXT', None)
            if vtxt is None:
                self._version = None
            else:
                vfile = open(vtxt, 'r')
                self._version = vfile.read()
                vfile.close()
        return self._version

router = Router()


def autodiscover():
    for app_name in settings.INSTALLED_APPS:
        app = import_module(app_name)
        if module_has_submodule(app, 'rest'):
            import_module(app_name + '.rest')
    for app_name in settings.INSTALLED_APPS:
        app = import_module(app_name)
        if module_has_submodule(app, 'serializers'):
            import_module(app_name + '.serializers')
    for app_name in settings.INSTALLED_APPS:
        app = import_module(app_name)
        if module_has_submodule(app, 'views'):
            import_module(app_name + '.views')



from PIL import Image, ImageOps
import StringIO
import os
from wq.io.util import guess_type
import subprocess
from django.conf import settings


def generate_image(image, size):
    size = int(size)
    path = os.path.join(settings.MEDIA_ROOT, image)
    name = os.path.basename(image)
    mime = guess_type(path)
    if mime.startswith('image/'):
        img = Image.open(path)
    elif mime.startswith('video/'):
        thumbdir = os.path.join(
            settings.MEDIA_ROOT, 'thumb', os.path.dirname(image)
        )
        try:
            os.makedirs(thumbdir)
        except OSError:
            pass
        thumbpath = os.path.join(thumbdir, name) + ".jpg"
        if not os.path.exists(thumbpath):
            subprocess.call(['avconv', '-i', path, '-vframes', '1', thumbpath])
        img = Image.open(thumbpath)
    else:
        return None

    if hasattr(img, '_getexif'):
        exif = img._getexif()
    else:
        exif = None
    if exif:
        orientation = exif.get(0x0112, 1)
    else:
        orientation = 1

    rotate = {
        3: Image.ROTATE_180,
        6: Image.ROTATE_270,
        8: Image.ROTATE_90
    }
    if orientation in rotate:
        img = img.transpose(rotate[orientation])

    if size < 256:
        img = ImageOps.fit(img, (size, size), Image.ANTIALIAS)
    elif size < 600:
        img.thumbnail((size, size), Image.ANTIALIAS)
    else:
        width, height = img.size
        if width > size:
            ratio = float(size) / float(width)
            height = int(height * ratio)
            img = img.resize((size, height), Image.ANTIALIAS)

    tdir = os.path.join(settings.MEDIA_ROOT, str(size), os.path.dirname(image))
    try:
        os.makedirs(tdir)
    except OSError:
        pass
    img.save(os.path.join(tdir, name), 'JPEG')
    data = StringIO.StringIO()
    img.save(data, 'JPEG')
    data.seek(0)
    return data



from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.contrib import admin

import swapper
from .models import AnnotatedModel
from .forms import AnnotationForm


class AnnotationInline(generic.GenericTabularInline):
    model = swapper.load_model('annotate', 'Annotation', required=False)
    form = AnnotationForm
    extra = 0

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == 'type':
            ctype = ContentType.objects.get_for_model(self.parent_model)
            AnnotationType = swapper.load_model(
                'annotate', 'AnnotationType', required=False
            )
            if AnnotationType:
                kwargs["queryset"] = AnnotationType.objects.filter(
                    contenttype=ctype
                )
        return super(AnnotationInline, self).formfield_for_foreignkey(
            db_field, request, **kwargs
        )


class AnnotatedModelAdmin(admin.ModelAdmin):
    model = AnnotatedModel
    inlines = [
        AnnotationInline,
    ]


class AnnotationTypeAdmin(admin.ModelAdmin):
    list_display = ('__unicode__', 'contenttype')
    list_filter = ('contenttype',)



from wq.db.patterns import models
import swapper
from django.db.models.signals import post_save
from django.core.exceptions import ImproperlyConfigured
from django.dispatch import receiver
from django.utils.timezone import now
from django.conf import settings
from collections import OrderedDict
from .compat import clone_field

swapper.set_app_prefix('vera', 'WQ')
MODELS = swapper.get_model_names(
    'vera', ('Site', 'Event', 'Report', 'ReportStatus', 'Parameter', 'Result')
)

VALID_REPORT_ORDER = getattr(settings, "WQ_VALID_REPORT_ORDER", ('-entered',))
# Base classes for Site-Event-Report-Attribute-Value pattern
# Extend these when swapping out default implementation (below)


class BaseSite(models.NaturalKeyModel):
    @property
    def valid_events(self):
        events = self.event_set.filter(
            report__status__is_valid=True
        ).values_list('pk', flat=True)
        # FIXME: events may be duplicated
        return self.event_set.filter(pk__in=events)

    class Meta:
        abstract = True


class BaseEvent(models.NaturalKeyModel):
    site = models.ForeignKey(MODELS['Site'], null=True, blank=True)

    @property
    def valid_reports(self):
        return self.report_set.filter(status__is_valid=True).order_by(
            *VALID_REPORT_ORDER
        )

    @property
    def vals(self):
        return OrderedDict([
            (a.type.natural_key()[0], a.value)
            for a in self.results
        ])

    # Valid results for this event
    @property
    def results(self):
        Result = swapper.load_model('vera', 'Result')
        return Result.objects.valid_results(report__event=self)

    @property
    def is_valid(self):
        return self.valid_reports.count() > 0

    class Meta:
        abstract = True


class ReportManager(models.RelatedModelManager):
    def create_report(self, event_key, values, **kwargs):
        Event = swapper.load_model('vera', 'Event')
        kwargs['event'] = Event.objects.find(*event_key)
        report = self.create(**kwargs)
        report.vals = values
        return report


class ValidReportManager(ReportManager):
    def get_queryset(self):
        qs = super(ValidReportManager, self)
        return qs.filter(status__is_valid=True).order_by(*VALID_REPORT_ORDER)


class BaseReport(models.RelatedModel):
    event = models.ForeignKey(MODELS['Event'])
    entered = models.DateTimeField(blank=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True,

        # Needed for model swapping in tests
        related_name="%(app_label)s_%(class)s",
    )
    status = models.ForeignKey(MODELS['ReportStatus'], null=True, blank=True)

    objects = ReportManager()
    valid_objects = ValidReportManager()

    @property
    def is_valid(self):
        return self.status and self.status.is_valid

    @property
    def vals(self):
        vals = OrderedDict()
        for result in self.results.all():
            vals[result.type.natural_key()[0]] = result.value
        return vals

    @vals.setter
    def vals(self, vals):
        Parameter = swapper.load_model('vera', 'Parameter')
        keys = [(key,) for key in vals.keys()]
        params, success = Parameter.objects.resolve_keys(keys)
        if not success:
            missing = [
                name for name, params in params.items() if params is None
            ]
            raise TypeError(
                "Could not identify one or more parameters: %s!"
                % missing
            )

        for key, param in params.items():
            result, is_new = self.results.get_or_create(type=param)
            result.value = vals[key[0]]
            result.save()

    def save(self, *args, **kwargs):
        if not self.entered:
            self.entered = now()
        super(BaseReport, self).save(*args, **kwargs)

    def __unicode__(self):
        if self.pk is not None:
            return "%s according to %s" % (self.event, self.user)
        else:
            return "New Report"

    class Meta:
        abstract = True
        ordering = VALID_REPORT_ORDER


class BaseReportStatus(models.NaturalKeyModel):
    slug = models.SlugField()
    name = models.CharField(max_length=255)
    is_valid = models.BooleanField(default=False)

    def __unicode__(self):
        return self.name

    class Meta:
        abstract = True
        unique_together = [['slug']]


class BaseParameter(models.IdentifiedRelatedModel):
    name = models.CharField(max_length=255)
    is_numeric = models.BooleanField(default=False)
    units = models.CharField(max_length=50, null=True, blank=True)

    def __unicode__(self):
        if self.units:
            return u"%s (%s)" % (self.name, self.units)
        else:
            return self.name

    class Meta:
        abstract = True
        ordering = ('name',)


class ResultManager(models.Manager):
    def valid_results(self, **filter):
        Parameter = swapper.load_model('vera', 'Parameter')

        filter['report__status__is_valid'] = True
        filter['empty'] = False

        # DISTINCT ON event, then parameter, collapsing results from different
        # reports into one
        distinct = ['report__event__id'] + nest_ordering(
            'type', Parameter._meta.ordering
        ) + ['type__id']

        # ORDER BY distinct fields, then valid report order
        order = distinct + nest_ordering('report', VALID_REPORT_ORDER)

        return self.filter(
            **filter
        ).order_by(*order).distinct(*distinct).select_related('report')


class BaseResult(models.Model):
    type = models.ForeignKey(MODELS['Parameter'])
    report = models.ForeignKey(MODELS['Report'], related_name='results')
    value_numeric = models.FloatField(null=True, blank=True)
    value_text = models.TextField(null=True, blank=True)
    empty = models.BooleanField(default=False, db_index=True)

    objects = ResultManager()

    def is_empty(self, value):
        if value is None:
            return True
        if isinstance(value, basestring) and len(value.strip()) == 0:
            return True
        return False

    @property
    def value(self):
        if self.type.is_numeric:
            return self.value_numeric
        return self.value_text

    @value.setter
    def value(self, val):
        self.empty = self.is_empty(val)
        if self.type.is_numeric:
            if self.empty:
                self.value_numeric = None
            else:
                self.value_numeric = val
        else:
            self.value_text = val

    def __unicode__(self):
        return "%s -> %s: %s" % (self.report, self.type, self.value)

    class Meta:
        abstract = True
        ordering = ('type',)
        index_together = [
            ('type', 'report', 'empty'),
        ]


# Denormalize Events with Results from valid Reports
class EventResultManager(models.Manager):
    def set_for_event(self, event, delete=True):
        """
        Update EventResult cache for the given event.
        """
        if delete:
            self.filter(event=event).delete()
        for result in event.results:
            er = self.model(
                event=event,
                result=result
            )
            er.save()

    def set_for_events(self, delete=True, **event_filter):
        """
        Update EventResult cache for the given events.  The event query should
        be specified as query keyword arguments, rather than a queryset, so
        that a JOIN can be used instead of retrieving the results for each
        event separately.
        """

        # Delete existing EventResults (using denormalized Event fields)
        if delete:
            er_filter = {
                'event_' + key: val for key, val in event_filter.items()
            }
            self.filter(**er_filter).delete()

        # Filter results (using JOIN through Report to Event)
        Result = swapper.load_model('vera', 'Result')
        result_filter = {
            'report__event__' + key: val for key, val in event_filter.items()
        }
        ers = []
        results = Result.objects.valid_results(
            **result_filter
        ).select_related('report__event')
        for result in results:
            er = self.model(
                event=result.report.event,
                result=result
            )
            er.denormalize()
            ers.append(er)
        self.bulk_create(ers)

    def set_all(self):
        """
        Reset the entire EventResult cache.
        """
        self.set_for_events()


class BaseEventResult(models.Model):
    """
    Denormalized event-result pairs (for valid reports)
    """
    id = models.PositiveIntegerField(primary_key=True)
    event = models.ForeignKey(MODELS['Event'])
    result = models.ForeignKey(MODELS['Result'])
    objects = EventResultManager()

    @property
    def result_value(self):
        if self.result_type.is_numeric:
            return self.result_value_numeric
        return self.result_value_text

    def __unicode__(self):
        return "%s -> %s: %s" % (
            self.event,
            self.result_type,
            self.result_value
        )

    def save(self, *args, **kwargs):
        if self.event_id is None or self.result_id is None:
            return
        self.denormalize()
        super(BaseEventResult, self).save(*args, **kwargs)

    def denormalize(self):
        """
        Denormalize all properties from the event and the result.
        """
        self.pk = self.result.pk

        def set_value(src, field):
            if field.primary_key:
                return
            name = field.name
            if field.rel:
                name += "_id"
            obj = getattr(self, src)
            setattr(self, src + '_' + name, getattr(obj, name))

        for field in self.event._meta.fields:
            set_value('event', field)
        for field in self.result._meta.fields:
            set_value('result', field)

    class Meta:
        abstract = True


def create_eventresult_model(event_cls, result_cls,
                             base=BaseEventResult, swappable=False):
    """
    **Here be magic**

    EventResult needs to have all of the fields from Event and Result.
    In order to DRY (and since either of these classes may be swapped), we
    add the fields dynamically here.  If neither Event or Result are swapped,
    this function will be called below, otherwise the user should call
    this function in their models.py.
    """

    app = 'vera'
    module = base.__module__
    for cls in event_cls, result_cls, base:
        if cls._meta.app_label != 'vera':
            app = cls._meta.app_label
            module = cls.__module__

    class Meta(base.Meta):
        db_table = 'wq_eventresult'
        unique_together = ('event', 'result_type')
        app_label = app

    if swappable:
        Meta.swappable = swappable

    attrs = {
        'Meta': Meta,
        '__module__': module
    }

    def add_field(prefix, field):
        if field.primary_key:
            return
        attrs[prefix + '_' + field.name] = clone_field(field)

    for field in event_cls._meta.fields:
        add_field('event', field)

    for field in result_cls._meta.fields:
        add_field('result', field)

    EventResult = type(
        'EventResult',
        (base,),
        attrs
    )

    # Disconnect any existing receiver
    post_save.disconnect(dispatch_uid="eventresult_receiver")

    @receiver(post_save, weak=False, dispatch_uid="eventresult_receiver")
    def handler(sender, instance=None, **kwargs):
        event = find_event(instance)
        if event:
            EventResult.objects.set_for_event(event)

    return EventResult


# Default implementation of the above classes, can be swapped
class Site(BaseSite):
    latitude = models.FloatField(null=True, blank=True)
    longitude = models.FloatField(null=True, blank=True)

    def __unicode__(self):
        return "%s, %s" % (round(self.latitude, 3), round(self.longitude, 3))

    class Meta(BaseSite.Meta):
        db_table = 'wq_site'
        swappable = swapper.swappable_setting('vera', 'Site')
        unique_together = ('latitude', 'longitude')


class Event(BaseEvent):
    date = models.DateField()

    def __unicode__(self):
        return "%s on %s" % (self.site, self.date)

    class Meta(BaseEvent.Meta):
        db_table = 'wq_event'
        swappable = swapper.swappable_setting('vera', 'Event')
        unique_together = ('site', 'date')
        ordering = ('-date',)


class Report(BaseReport):
    class Meta(BaseReport.Meta):
        db_table = 'wq_report'
        swappable = swapper.swappable_setting('vera', 'Report')


class ReportStatus(BaseReportStatus):
    class Meta(BaseReportStatus.Meta):
        verbose_name_plural = 'report statuses'
        db_table = 'wq_reportstatus'
        swappable = swapper.swappable_setting('vera', 'ReportStatus')


class Parameter(BaseParameter):
    class Meta(BaseParameter.Meta):
        db_table = 'wq_parameter'
        swappable = swapper.swappable_setting('vera', 'Parameter')


class Result(BaseResult):
    class Meta(BaseResult.Meta):
        db_table = 'wq_result'
        swappable = swapper.swappable_setting('vera', 'Result')


EventResult = create_eventresult_model(
    Event, Result, swappable=swapper.swappable_setting('vera', 'EventResult')
)

if ((swapper.is_swapped('vera', 'Event')
        or swapper.is_swapped('vera', 'Result'))
        and not swapper.is_swapped('vera', 'EventResult')):
    raise ImproperlyConfigured(
        "Event or Result was swapped but EventResult was not!"
    )


def nest_ordering(prefix, ordering, ignore_reverse=False):
    nest_order = []
    for field in ordering:
        reverse = ''
        if field.startswith('-'):
            field = field[1:]
            if not ignore_reverse:
                reverse = '-'
        nest_order.append(reverse + prefix + '__' + field)
    return nest_order


def find_event(instance):
    if isinstance(instance, BaseEvent):
        return instance
    if isinstance(instance, BaseReport):
        return instance.event
    if isinstance(instance, BaseResult):
        return instance.report.event
    return None



from django.db import models
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic

from django.conf import settings
import swapper
from wq.db.patterns.models import AnnotatedModel
from wq.db.contrib.dbio.models import IoModel

from wq.io.util import guess_type

swapper.set_app_prefix('files', 'WQ')


# Custom FileField handles both images and files
class FileField(models.ImageField):
    # Use base forms.FileField to skip ImageField validation
    def formfield(self, **kwargs):
        kwargs['form_class'] = forms.FileField
        return super(FileField, self).formfield(**kwargs)

    # Only update_dimension_fields for images
    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        if (getattr(instance, 'mimetype', None) is not None
                and 'image' in instance.mimetype):
            super(FileField, self).update_dimension_fields(instance, force,
                                                           *args, **kwargs)
        else:
            pass

    # Allow model to specify upload directory
    def generate_filename(self, instance, filename):
        if hasattr(instance, 'get_directory'):
            self.upload_to = instance.get_directory()
        return super(FileField, self).generate_filename(instance, filename)


class FileType(models.Model):
    name = models.CharField(max_length=255)
    mimetype = models.CharField(max_length=255)

    def __unicode__(self):
        return '%s (%s)' % (self.name, self.mimetype)

    class Meta:
        db_table = 'wq_filetype'


class FileManager(models.Manager):
    def get_query_set(self):
        qs = super(FileManager, self).get_query_set()
        if self.model.type_name is not None:
            return qs.filter(type__name=self.model.type_name)
        else:
            return qs


class BaseFile(AnnotatedModel, IoModel):
    name = models.CharField(max_length=255, null=True, blank=True)
    type = models.ForeignKey(FileType, null=True, blank=True)
    file = FileField(upload_to='.', width_field='width', height_field='height')
    size = models.IntegerField(null=True, blank=True)
    width = models.IntegerField(null=True, blank=True)
    height = models.IntegerField(null=True, blank=True)

    type_name = "File"

    def get_directory(self):
        if self.is_image:
            return 'images'
        else:
            return 'files'

    @property
    def mimetype(self):
        if self.type is not None:
            return self.type.mimetype

        if self.file.name not in (None, ""):
            self.file.open()
            mimetype = guess_type(self.file.name, self.file.read(1024))
            self.file.close()
            return mimetype
        else:
            return None

    @property
    def is_image(self):
        return self.mimetype is not None and self.mimetype.startswith('image/')

    def load_io(self, **options):
        from wq.io import load_file
        filename = "%s/%s" % (settings.MEDIA_ROOT, self.file.name)
        if not options:
            options = self.load_file_options()
        return load_file(filename, options=options)

    def load_file_options(self):
        headers = self.relationships.filter(
            type__name='Contains Column',
            range__type='list'
        )
        if headers.exists():
            header_row = headers[0].range_set.get(type='head').start_row
            start_row = headers[0].range_set.get(type='list').start_row
            return {
                'header_row': header_row,
                'start_row': start_row
            }

        templates = self.inverserelationships.filter(
            type__inverse_name='Template'
        )
        if templates.exists():
            template = templates[0].right
            return template.load_file_options()
        return {}

    def save(self, *args, **kwargs):
        if self.type is None:
            self.type, isnew = FileType.objects.get_or_create(
                mimetype=self.mimetype,
                name=self.type_name
            )
        if self.size is None:
            self.size = self.file.size
        if self.name is None or self.name == '':
            self.name = self.file.name
        super(BaseFile, self).save(*args, **kwargs)

    def __unicode__(self):
        if self.name is not None:
            return self.name
        else:
            return 'File %s' % self.id

    class Meta:
        abstract = True


class File(BaseFile):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True)

    class Meta:
        db_table = 'wq_file'
        ordering = ("name",)
        swappable = swapper.swappable_setting('files', 'File')

# Tell south not to worry about the "custom" field type
from south.modelsinspector import add_introspection_rules
add_introspection_rules([], ["^wq.db.contrib.files.models.FileField"])



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models
import swapper


class Migration(SchemaMigration):

    def forwards(self, orm):

        if not swapper.is_swapped('annotate', 'AnnotationType'):

            # Adding model 'AnnotationType'
            db.create_table('wq_annotationtype', (
                ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
                ('contenttype', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'], null=True, blank=True)),
            ))
            db.send_create_signal('annotate', ['AnnotationType'])

        if not swapper.is_swapped('annotate', 'Annotation'):
            AnnotationType = swapper.load_model('annotate', 'AnnotationType', orm)

            # Adding model 'Annotation'
            db.create_table('wq_annotation', (
                ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),

                ('type', self.gf('django.db.models.fields.related.ForeignKey')(to=AnnotationType)),
                ('value', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
                ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'])),
                ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
            ))
            db.send_create_signal('annotate', ['Annotation'])

            # Create index
            db.execute("""CREATE INDEX wq_annotation_idx ON wq_annotation
                             (content_type_id, object_id)""")

    def backwards(self, orm):

        if not swapper.is_swapped('annotate', 'AnnotationType'):
            # Deleting model 'AnnotationType'
            db.delete_table('wq_annotationtype')

        if not swapper.is_swapped('annotate', 'Annotation'):
            # Drop index
            db.execute("DROP INDEX wq_annotation_idx;")

            # Deleting model 'Annotation'
            db.delete_table('wq_annotation')

    models = {
        'annotate.annotation': {
            'Meta': {'object_name': 'Annotation', 'db_table': "'wq_annotation'"},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['annotate.AnnotationType']"}),
            'value': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        'annotate.annotationtype': {
            'Meta': {'object_name': 'AnnotationType', 'db_table': "'wq_annotationtype'"},
            'contenttype': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']", 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['annotate']



from wq.db.rest import app
from .serializers import FileSerializer
from .models import FileType
import swapper
File = swapper.load_model('files', 'File')

app.router.register_model(FileType)
app.router.register_model(File, serializer=FileSerializer)



from rest_framework import serializers
from wq.db.patterns.base.serializers import TypedAttachmentSerializer
from .models import Identifier, Authority


class IdentifierSerializer(TypedAttachmentSerializer):
    type_model = Authority
    type_field = 'authority'
    attachment_fields = ['id', 'name', 'slug', 'is_primary']
    required_fields = ['name']

    url = serializers.Field()

    @property
    def expected_types(self):
        return [None] + list(Authority.objects.all())

    def create_dict(self, atype, data, fields, index):
        obj = super(IdentifierSerializer, self).create_dict(
            atype, data, fields, index
        )
        if obj is None:
            return obj
        if 'is_primary' not in obj:
            obj['is_primary'] = (index == 0)
        return obj

    class Meta(TypedAttachmentSerializer.Meta):
        read_only_fields = ('slug',)



from rest_framework.filters import BaseFilterBackend
RESERVED_PARAMETERS = ('_', 'page', 'limit', 'format', 'slug', 'mode')

from .models import get_ct


class FilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        kwargs = {}
        for key, val in view.kwargs.items() + request.GET.items():
            if key in RESERVED_PARAMETERS or key in view.ignore_kwargs:
                continue
            kwargs[key] = val if isinstance(val, unicode) else val[0]
        model = getattr(view, 'model', None) or queryset.model
        ctype = get_ct(model)

        for key, val in kwargs.items():
            if key in ('target',):
                del kwargs[key]
                continue
            found = False
            for f in model._meta.fields:
                if ((f.name != key and ctype.identifier + f.name != key)
                        or f.name == ctype.model):
                    continue
                found = True
                if getattr(f, 'rel', None):
                    del kwargs[key]
                    pcls = f.rel.to
                    router = getattr(view, 'router', None)
                    if router:
                        slug = router.get_lookup_for_model(pcls)
                        kwargs[f.name] = pcls.objects.get(**{slug: val})
                    elif get_ct(f.rel.to).is_identified:
                        kwargs[f.name] = pcls.objects.get_by_identifier(val)
                    else:
                        kwargs[f.name] = pcls.objects.get(pk=val)

            if not found and ctype.is_related:
                for pct in ctype.get_all_parents():
                    if pct.identifier == key:
                        pclass = pct.model_class()
                        if pct.is_identified:
                            parent = pclass.objects.get_by_identifier(
                                kwargs[key]
                            )
                        else:
                            parent = pclass.objects.get(pk=kwargs[key])
                        del kwargs[key]
                        objs = model.objects.filter_by_related(parent)
                        kwargs['pk__in'] = objs.values_list('pk', flat=True)

        return queryset.filter(**kwargs)



from wq.db.patterns import models


class MetaColumn(models.IdentifiedRelatedModel):
    DATA_TYPES = (
        ('site', 'Site Metadata'),
        ('event', 'Event Metadata'),
        ('report', 'Report Metadata'),
        ('parameter', 'Parameter Metadata'),
        ('result', 'Result Data/Metadata'),
    )

    name = models.CharField(max_length=255)
    type = models.CharField(max_length=10, choices=DATA_TYPES, blank=True)

    def __unicode__(self):
        name = super(MetaColumn, self).__unicode__()
        return "%s (%s)" % (name, self.get_type_display())

    class Meta:
        db_table = 'wq_metacolumn'


class UnknownItem(models.IdentifiedRelatedModel):
    name = models.CharField(max_length=255)

    def __unicode__(self):
        return self.name

    class Meta:
        db_table = 'wq_unknownitem'


class SkippedRecord(models.Model):
    reason = models.TextField()

    def __unicode__(self):
        return self.reason

    class Meta:
        db_table = 'wq_skippedrecord'


class Range(models.Model):
    RANGE_TYPES = (
        ('head', 'Column header / label'),
        ('value', 'Global Value'),
        ('list', 'Data series'),
        ('row', 'Individual Record'),
    )
    relationship = models.ForeignKey(models.Relationship)
    type = models.CharField(max_length=10, choices=RANGE_TYPES)
    start_row = models.IntegerField()
    end_row = models.IntegerField(null=True, blank=True)
    start_column = models.IntegerField()
    end_column = models.IntegerField()

    def __unicode__(self):
        if self.start_row == self.end_row:
            row = "Row %s" % self.start_row
        else:
            if self.end_row is not None:
                row = "Rows %s-%s" % (self.start_row, self.end_row)
            else:
                row = "Row %s onward" % (self.start_row,)

        if self.start_column == self.end_column:
            column = "Column %s" % self.start_column
        else:
            column = "Column %s-%s" % (self.start_column, self.end_column)

        return "%s at %s, %s" % (self.relationship, row, column)

    class Meta:
        db_table = 'wq_range'


class IoModel(models.RelatedModel):
    def load_io(self, **options):
        raise NotImplementedError()

    def already_parsed(self):
        return self.relationships.filter(
            type__name='Contains Column',
            range__type='list'
        ).exists()

    def get_id_choices(self, model, meta):
        return model.objects.all()

    class Meta:
        abstract = True



from rest_framework.serializers import FileField as RestFileField, Field
from wq.db.rest import serializers
from .models import FileField


class FileSerializer(serializers.ModelSerializer):
    is_image = Field('is_image')

    def __init__(self, *args, **kwargs):
        self.field_mapping[FileField] = RestFileField
        super(FileSerializer, self).__init__(*args, **kwargs)

    def from_native(self, data, files):
        obj = super(FileSerializer, self).from_native(data, files)
        if 'request' in self.context and obj and obj.user is None:
            user = self.context['request'].user
            if user.is_authenticated():
                obj.user = user
        return obj



from wq.db.rest.serializers import ModelSerializer
from wq.db.rest.models import get_ct, get_object_id


class TypedAttachmentSerializer(ModelSerializer):
    attachment_fields = ['id']
    required_fields = []
    type_model = None
    type_field = 'type'
    object_field = 'content_object'

    def to_native(self, obj):
        data = super(TypedAttachmentSerializer, self).to_native(obj)

        has_parent = self.parent and hasattr(self.parent.opts, 'model')
        if has_parent:
            # This is being serialized with its parent object, don't need to
            # reference the parent again.
            pass
        else:
            # Include pointer to parent object. Collapse content_type and
            # object_id into a single attribute: the identifier for the content
            # type plus '_id' e.g. if contenttype's name is 'visit' then the
            # attribute is 'visit_id'.  This lets us pretend the generic
            # foreign key is a regular one in client.  For the value, Use
            # get_object_id insted of obj.object_id, in case the the content
            # object is an IdentifiedModel

            parent_obj = getattr(obj, self.object_field)
            idname = get_ct(parent_obj).identifier + '_id'
            data[idname] = get_object_id(parent_obj)
        return data

    @property
    def expected_types(self):
        return self.type_model.objects.all()

    def create_dict(self, atype, data, fields, index):
        attachment = {
            self.type_field: atype.pk if atype else None,
        }
        for field in fields:
            if fields[field] in data:
                attachment[field] = data[fields[field]]
                if field == 'id' and attachment[field]:
                    attachment[field] = int(attachment[field])
        for field in self.required_fields:
            if not attachment.get(field, None):
                return None
        return attachment

    def field_from_native(self, data, files, field_name, into):
        # Handle attachments that are submitted together with their parent

        # Ideal case: an array of dicts, this can be handled by the default
        # implementation
        if field_name in data:
            return super(TypedAttachmentSerializer, self).field_from_native(
                data, files, field_name, into
            )
        # Common case: form submission.  In this case each attachment should
        # be submitted as form fields with names in the format [model]_[typeid]
        # e.g. annotation_23=30.5 will become {'type': 23, 'value': '30.5'}

        # Retrieve form values into more usable array/dict format
        attachments = []

        ct = get_ct(self.opts.model)
        i = 0
        for atype in self.expected_types:
            fields = {
                afield: '%s-%s-%s' % (
                    ct.identifier,
                    get_object_id(atype) if atype else '',
                    afield
                ) for afield in self.attachment_fields
            }
            found = False
            for key in fields.values():
                if key in data:
                    found = True
            if found:
                attachment = self.create_dict(atype, data, fields, i)
                if attachment:
                    attachments.append(attachment)
                i += 1

        # Send modified object to default implementation
        return super(TypedAttachmentSerializer, self).field_from_native(
            {field_name: attachments}, files, field_name, into
        )

    class Meta:
        # Don't validate these fields (items are saved with their parent)
        exclude = ("content_type", "content_type_id",
                   "object_id", "label", "for")



from django.conf.urls import patterns, include, url
from rest_framework.urlpatterns import format_suffix_patterns
from .views import TimeSeriesView, ScatterView, BoxPlotView

ids = r'^(?P<ids>[^\.]+)'

urlpatterns = patterns('',
    url(ids + r'/timeseries$', TimeSeriesView.as_view()),
    url(ids + r'/scatter$', ScatterView.as_view()),
    url(ids + r'/boxplot$', BoxPlotView.as_view()),
)
urlpatterns = format_suffix_patterns(urlpatterns)



from wq.db.rest import app
import swapper
from .serializers import EventSerializer, ReportSerializer, ResultSerializer

Site = swapper.load_model('vera', 'Site')
Event = swapper.load_model('vera', 'Event')
Report = swapper.load_model('vera', 'Report')
ReportStatus = swapper.load_model('vera', 'ReportStatus')
Parameter = swapper.load_model('vera', 'Parameter')
Result = swapper.load_model('vera', 'Result')

app.router.register_model(Site)
app.router.register_model(Event, serializer=EventSerializer)
app.router.register_model(Report, serializer=ReportSerializer)
app.router.register_model(ReportStatus, lookup='slug')
app.router.register_model(Parameter)
app.router.register_model(Result, serializer=ResultSerializer)



from wq.db.rest.models import MultiQuerySet
from wq.db.patterns.models import Identifier
import swapper

if swapper.is_swapped('annotate', 'Annotation'):
    Annotation = None
else:
    Annotation = swapper.load_model('annotate', 'Annotation')


def search(query, auto=True, content_type=None, authority_id=None):
    if content_type:
        ctfilter = {'content_type__model': content_type}
    else:
        ctfilter = {}

    idfilter = ctfilter.copy()
    if authority_id:
        idfilter['authority_id'] = authority_id

    distinct_on = ('content_type__id', 'object_id')
    # First check for exact identifier matches
    id_matches = Identifier.objects.filter_by_identifier(query)
    id_matches = id_matches.filter(**idfilter)
    # If "auto" mode and only one distinct object, return first identifier
    if id_matches.distinct(*distinct_on).count() == 1 and auto:
        return id_matches[0:1]

    # Otherwise, include any identifiers containing the case-insensitive string
    id_matches = id_matches | Identifier.objects.filter(
        name__icontains=query,
        **idfilter
    ) | Identifier.objects.filter(
        slug__icontains=query,
        **idfilter
    )

    # Avoid duplicates due to multiple similar ids on the same object
    id_matches = id_matches.order_by(*distinct_on).distinct(*distinct_on)

    # Then, include any annotations containing the string (case insensitive)
    results = [id_matches]
    if Annotation:
        annot_matches = Annotation.objects.filter(
            value__icontains=query
        ).filter(**ctfilter)

        if content_type:
            # If (and only if) content type filter is specified, we can also
            # guarantee that annot_matches won't contain objects already found
            # in id_matches
            annot_matches = annot_matches.exclude(
                object_id__in=id_matches.values_list('object_id', flat=True)
            )
        results.append(annot_matches)

    return MultiQuerySet(*results)



from wq.db.patterns.base.serializers import TypedAttachmentSerializer
import swapper

from wq.db.rest.models import get_ct

AnnotationType = swapper.load_model('annotate', 'AnnotationType')


class AnnotationSerializer(TypedAttachmentSerializer):
    attachment_fields = ['id', 'value']
    type_model = AnnotationType

    @property
    def expected_types(self):
        ct = get_ct(self.parent.opts.model)
        return AnnotationType.objects.filter(contenttype=ct)



from django.conf.urls import patterns, url

from wq.db.rest import app
from .views import GenerateView

urlpatterns = patterns('',
    url(r'^generate/(?P<size>\d+)/(?P<image>.+)$', GenerateView.as_view()),
)



from wq.db.rest.serializers import ModelSerializer, ContentTypeField
from wq.db.patterns.base.serializers import TypedAttachmentSerializer
from wq.db.rest.models import get_ct, get_object_id, get_by_identifier

from .models import (
    Relationship, InverseRelationship,
    RelationshipType, InverseRelationshipType
)


class RelationshipSerializer(TypedAttachmentSerializer):
    attachment_fields = ['id', 'item_id']
    required_fields = ['item_id']
    type_model = RelationshipType
    object_field = 'left'
    item_id_field = 'to_object_id'
    parent_id_field = 'from_object_id'

    def get_default_fields(self):
        fields = super(RelationshipSerializer, self).get_default_fields()
        del fields[self.parent_id_field]
        fields['type_label'].source = 'reltype'
        return fields

    def to_native(self, rel):
        data = super(RelationshipSerializer, self).to_native(rel)
        del data[self.item_id_field]
        data.update(rel.right_dict)
        return data

    def create_dict(self, atype, data, fields, index):
        attachment = super(RelationshipSerializer, self).create_dict(
            atype, data, fields, index
        )
        if attachment is None:
            return None
        if 'item_id' in attachment and 'type' in attachment:
            type = self.type_model.objects.get(pk=attachment['type'])
            cls = type.right.model_class()
            obj = get_by_identifier(cls.objects, attachment['item_id'])
            attachment[self.item_id_field] = obj.pk
            del attachment['item_id']
        return attachment


class InverseRelationshipSerializer(RelationshipSerializer):
    item_id_field = 'from_object_id'
    parent_id_field = 'to_object_id'
    type_model = InverseRelationshipType

    def get_default_fields(self):
        fields = super(
            InverseRelationshipSerializer, self
        ).get_default_fields()
        return fields


class RelationshipTypeSerializer(ModelSerializer):
    from_type = ContentTypeField()
    to_type = ContentTypeField()

    class Meta:
        exclude = ('for',)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from wq.db.patterns.base import SerializableGenericRelation
from wq.db.patterns.base.models import NaturalKeyModel
from django.contrib import admin
from django.core.exceptions import FieldError
from collections import OrderedDict
import swapper

swapper.set_app_prefix('annotate', 'WQ')
ANNOTATIONTYPE_MODEL = swapper.get_model_name('annotate', 'AnnotationType')
ANNOTATION_MODEL = swapper.get_model_name('annotate', 'Annotation')


class BaseAnnotationType(NaturalKeyModel):
    name = models.CharField(max_length=255)
    contenttype = models.ForeignKey(ContentType, null=True, blank=True)

    # Assign a value to annotated_model to automatically set contenttype
    # (Useful for subclasses with a single target model)
    annotated_model = None

    def __unicode__(self):
        for f in self._meta.get_all_related_objects():
            # Check for any linked subclasses as they may have a more
            # meaningful representation
            if f.field.__class__ == models.OneToOneField:
                try:
                    child = getattr(self, f.get_accessor_name())
                except f.field.model.DoesNotExist:
                    continue
                else:
                    return unicode(child)
        # Fall back to name
        return self.name

    def natural_key(self):
        return (self.name,)

    def clean(self, *args, **kwargs):
        if self.annotated_model is not None:
            self.contenttype = ContentType.objects.get_for_model(
                self.annotated_model
            )

    class Meta:
        abstract = True
        unique_together = [['name']]


class AnnotationType(BaseAnnotationType):
    class Meta(BaseAnnotationType.Meta):
        db_table = 'wq_annotationtype'
        swappable = swapper.swappable_setting('annotate', 'AnnotationType')


class AnnotationManager(models.Manager):

    # Collapse annotations into dict for simple access
    @property
    def vals(self):
        if (hasattr(self, '_vals')):
            return getattr(self, '_vals')

        vals = OrderedDict()
        for annot in self.all():
            vals[annot.type.natural_key()[0]] = annot.value

        setattr(self, '_vals', vals)
        return vals

    # Default implementation of get_or_create doesn't work well with generics
    def get_or_create(self, **kwargs):
        try:
            return self.get(**kwargs), False
        except self.model.DoesNotExist:
            return self.create(**kwargs), True

    # Fix filtering across generic relations
    def filter(self, *args, **kwargs):

        def add_content_type(model):
            ctype = ContentType.objects.get_for_model(model)
            if 'content_type' in kwargs and kwargs['content_type'] != ctype:
                raise FieldError(
                    "Cannot match more than one generic relationship!"
                )
            kwargs['content_type'] = ctype

        for rel in self.model._meta.get_all_related_many_to_many_objects():
            if not isinstance(rel.field, generic.GenericRelation):
                continue
            rname = rel.field.related_query_name()
            for key in kwargs.keys():
                if key == rname or key.startswith(rname + '__'):
                    add_content_type(rel.model)

        return super(AnnotationManager, self).filter(*args, **kwargs)


class BaseAnnotation(models.Model):
    type = models.ForeignKey(ANNOTATIONTYPE_MODEL)

    # Link can contain a pointer to any model
    # FIXME: restrict to models allowed for given type
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = generic.GenericForeignKey()

    objects = AnnotationManager()

    def __unicode__(self):
        return (
            '%(object)s -> %(annot)s: %(value)s'
            % {
                'object': self.content_object,
                'annot': self.type,
                'value': self.value
            })

    class Meta:
        abstract = True


class Annotation(BaseAnnotation):
    value = models.TextField(null=True, blank=True)

    class Meta:
        db_table = 'wq_annotation'
        swappable = swapper.swappable_setting('annotate', 'Annotation')


class AnnotationSet(SerializableGenericRelation):
    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            kwargs['to'] = ANNOTATION_MODEL
        kwargs['related_name'] = None
        super(AnnotationSet, self).__init__(*args, **kwargs)


class AnnotatedModel(models.Model):
    annotations = AnnotationSet()

    @property
    def vals(self):
        return self.annotations.vals

    @vals.setter
    def vals(self, vals):
        AnnotationType = swapper.load_model('annotate', 'AnnotationType')
        keys = [(key,) for key in vals.keys()]
        types, success = AnnotationType.objects.resolve_keys(keys)
        if not success:
            missing = [name for name, atype in types.items() if atype is None]
            raise TypeError(
                "Could not identify one or more annotation types: %s!"
                % missing
            )

        for name, atype in types.items():
            annot, is_new = self.annotations.get_or_create(type=atype)
            annot.value = vals[name[0]]
            annot.save()

    class Meta:
        abstract = True

# Tell south not to worry about the "custom" field type
from south.modelsinspector import add_ignored_fields
add_ignored_fields(["^wq.db.patterns.annotate.models.AnnotationSet"])



from rest_framework.response import Response
from wq.db.rest.views import SimpleView
from wq.db.rest.renderers import binary_renderer
from .util import generate_image


class GenerateView(SimpleView):
    renderer_classes = [binary_renderer('image/jpeg')]

    def get(self, request, size, image):
        data = generate_image(image, size)
        return Response(data.read())



from django.contrib.admin import *
from .annotate.admin import *
from .identify.admin import *
from .locate.admin import *
from .relate.admin import *

from .models import IdentifiedRelatedModel


class IdentifiedRelatedModelAdmin(IdentifiedModelAdmin, RelatedModelAdmin):
    model = IdentifiedRelatedModel
    inlines = [
        IdentifierInline,
        RelationshipInline,
        InverseRelationshipInline
    ]



from wq.db.rest import app
import swapper
from .serializers import AnnotationSerializer

Annotation = swapper.load_model('annotate', 'Annotation')
AnnotationType = swapper.load_model('annotate', 'AnnotationType')
app.router.register_model(Annotation, serializer=AnnotationSerializer)
app.router.register_model(AnnotationType)



from django.http import Http404
from rest_framework.generics import GenericAPIView as RestGenericAPIView
from rest_framework.response import Response
from rest_framework.decorators import link
from rest_framework import status, viewsets
from .models import get_ct, get_object_id, get_by_identifier
from django.conf import settings


class GenericAPIView(RestGenericAPIView):
    router = None
    ignore_kwargs = []

    @property
    def template_name(self):
        """
        Infer template name from view/viewset name
        """
        name = type(self).__name__
        name = name.replace('ViewSet', '')
        name = name.replace('View', '')
        return name.lower() + '.html'

    @property
    def depth(self):
        return 0

    def get_template_names(self):
        return [self.template_name]

    def get_queryset(self):
        if self.router is not None and self.model is not None:
            return self.router.get_queryset_for_model(self.model, self.request)
        return super(GenericAPIView, self).get_queryset()

    def get_serializer_class(self):
        if self.router is not None and self.model is not None:
            return self.router.get_serializer_for_model(self.model, self.depth)
        return super(GenericAPIView, self).get_serializer_class()

    def get_paginate_by(self):
        if self.paginate_by_param not in self.request.GET:
            if self.router is not None and self.model is not None:
                return self.router.get_paginate_by_for_model(self.model)
        return super(GenericAPIView, self).get_paginate_by()


class SimpleView(GenericAPIView):
    def get(self, request, *args, **kwargs):
        return Response({})


class SimpleViewSet(viewsets.ViewSet, GenericAPIView):
    def list(self, request, *args, **kwargs):
        return Response({})


class ModelViewSet(viewsets.ModelViewSet, GenericAPIView):
    target = None

    @property
    def template_name(self):
        basename = get_ct(self.model).identifier
        if self.action in ('retrieve', 'create', 'update', 'delete'):
            suffix = 'detail'
        else:
            suffix = self.action
        return "%s_%s.html" % (basename, suffix)

    @property
    def depth(self):
        if self.action in ('retrieve', 'edit'):
            return 1
        else:
            return 0

    @link()
    def edit(self, request, *args, **kwargs):
        """
        Generates a context appropriate for editing a form
        """
        response = self.retrieve(request, *args, **kwargs)
        self.add_lookups(response.data)
        return response

    def new(self, request):
        """
        new is a variant of the "edit" action, but with no existing model
        to lookup.
        """
        self.action = 'edit'
        init = request.GET.dict()
        for arg in self.ignore_kwargs:
            init.pop(arg, None)
        obj = self.model(**init)
        serializer = self.get_serializer(obj)
        data = serializer.data
        self.add_lookups(data)
        return Response(data)

    def retrieve(self, request, *args, **kwargs):
        """
        Custom retrieve watches for "new" lookup value and switches modes
        accordingly
        """
        if hasattr(self, 'lookup_url_kwarg'):
            # New in DRF 2.4?
            lookup = self.lookup_url_kwarg or self.lookup_field
        else:
            lookup = self.lookup_field

        if self.kwargs.get(lookup, "") == "new":
            # new/edit mode
            return self.new(request)
        else:
            # Normal detail view
            return super(ModelViewSet, self).retrieve(request, *args, **kwargs)

    def add_lookups(self, context):
        # Mimic _addLookups in wq.app/app.js
        context['edit'] = True
        ct = get_ct(self.model)
        for pct, fields in ct.get_foreign_keys().items():
            if not pct.is_registered():
                continue
            for field in fields:
                choices = self.get_lookup_choices(pct, context, field)
                self.set_selected(choices, context.get(field + '_id', ''))
                if field == pct.model:
                    context[pct.urlbase] = choices
                context[field + '_list'] = choices

    def set_selected(self, choices, value):
        for choice in choices:
            if choice['id'] == value:
                choice['selected'] = True

    def get_lookup_choices(self, ct, context, field_name=None):
        from wq.db.rest import app
        parent_model = ct.model_class()
        if not field_name:
            field_name = ct.model
        qs = app.router.get_queryset_for_model(parent_model)
        fn = getattr(self, 'get_%s_choices' % field_name, None)
        if fn:
            qs = fn(qs, context)
        return app.router.serialize(qs, many=True)

    def get_object(self, queryset=None):
        try:
            obj = super(ModelViewSet, self).get_object(queryset)
        except Http404:
            if not get_ct(self.model).is_identified:
                raise

            # Allow retrieval via non-primary identifiers
            slug = self.kwargs.get(self.slug_url_kwarg)
            try:
                obj = self.model.objects.get_by_identifier(slug)
            except:
                raise Http404("Could not find %s with id '%s'" % (
                    self.model._meta.verbose_name,
                    slug
                ))
            #TODO: automatically redirect to primary identifier?
        return obj

    def list(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).list(
            request, *args, **kwargs
        )
        if not isinstance(response.data, dict):
            return response

        if self.target:
            response.data['target'] = self.target
        ct = get_ct(self.model)
        for pct in get_ct(self.model).get_all_parents():
            self.get_parent(pct, response)
        return response

    def create(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).create(
            request, *args, **kwargs
        )
        if not request.accepted_media_type.startswith('text/html'):
            # JSON request, assume client will handle redirect
            return response

        # HTML request, probably a form post from an older browser
        if response.status_code == status.HTTP_201_CREATED:
            ct = get_ct(self.model)
            conf = ct.get_config(request.user)

            # Redirect to detail view
            postsave = conf.get('postsave', ct.identifier)
            if postsave != ct.identifier:
                # Optional: return to detail view of a parent model
                ct = get_ct(postsave)
                oid = response.data.get(postsave + '_id', None) or ""
            else:
                # Default: return to detail view of the saved model
                oid = response.data['id']

            url = '/%s/%s' % (ct.urlbase, oid)
            return Response(
                {'detail': 'Created'},
                status=status.HTTP_302_FOUND,
                headers={'Location': url}
            )
        else:
            errors = [{
                'field': key,
                'errors': val
            } for key, val in response.data.items()]
            template = get_ct(self.model).identifier + '_error.html'
            return Response(
                {
                    'errors': errors,
                    'post': request.DATA
                },
                status=response.status_code,
                template_name=template
            )

    def get_parent(self, ct, response):
        pid = self.kwargs.get(ct.identifier, None)
        if not pid:
            return

        pcls = ct.model_class()
        if self.router:
            slug = self.router.get_lookup_for_model(pcls)
            parent = pcls.objects.get(**{slug: pid})
        else:
            parent = get_by_identifier(pcls.objects, pid)
        if ct.urlbase == '':
            urlbase = ''
        else:
            urlbase = ct.urlbase + '/'
        objid = get_object_id(parent)
        response.data['parent_label'] = unicode(parent)
        response.data['parent_id'] = objid
        response.data['parent_url'] = '%s%s' % (urlbase, objid)
        response.data['parent_is_' + ct.identifier] = True
        if self.router:
            response.data['parent'] = self.router.serialize(parent)
        return parent



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'MetaColumn'
        db.create_table('wq_metacolumn', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('type', self.gf('django.db.models.fields.CharField')(max_length=10)),
        ))
        db.send_create_signal('dbio', ['MetaColumn'])

        # Adding model 'UnknownItem'
        db.create_table('wq_unknownitem', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
        ))
        db.send_create_signal('dbio', ['UnknownItem'])

        # Adding model 'SkippedRecord'
        db.create_table('wq_skippedrecord', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('reason', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal('dbio', ['SkippedRecord'])

        # Adding model 'Range'
        db.create_table('wq_range', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('relationship', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['relate.Relationship'])),
            ('type', self.gf('django.db.models.fields.CharField')(max_length=10)),
            ('start_row', self.gf('django.db.models.fields.IntegerField')()),
            ('end_row', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
            ('start_column', self.gf('django.db.models.fields.IntegerField')()),
            ('end_column', self.gf('django.db.models.fields.IntegerField')()),
        ))
        db.send_create_signal('dbio', ['Range'])

    def backwards(self, orm):
        # Deleting model 'MetaColumn'
        db.delete_table('wq_metacolumn')

        # Deleting model 'UnknownItem'
        db.delete_table('wq_unknownitem')

        # Deleting model 'SkippedRecord'
        db.delete_table('wq_skippedrecord')

        # Deleting model 'Range'
        db.delete_table('wq_range')

    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'identify.authority': {
            'Meta': {'object_name': 'Authority', 'db_table': "'wq_identifiertype'"},
            'homepage': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'object_url': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        u'identify.identifier': {
            'Meta': {'object_name': 'Identifier', 'db_table': "'wq_identifier'"},
            'authority': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['identify.Authority']", 'null': 'True', 'blank': 'True'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_primary': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'db_index': 'True'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'dbio.metacolumn': {
            'Meta': {'object_name': 'MetaColumn', 'db_table': "'wq_metacolumn'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'type': ('django.db.models.fields.CharField', [], {'max_length': '10'})
        },
        u'dbio.range': {
            'Meta': {'object_name': 'Range', 'db_table': "'wq_range'"},
            'end_column': ('django.db.models.fields.IntegerField', [], {}),
            'end_row': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'relationship': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['relate.Relationship']"}),
            'start_column': ('django.db.models.fields.IntegerField', [], {}),
            'start_row': ('django.db.models.fields.IntegerField', [], {}),
            'type': ('django.db.models.fields.CharField', [], {'max_length': '10'})
        },
        u'dbio.skippedrecord': {
            'Meta': {'object_name': 'SkippedRecord', 'db_table': "'wq_skippedrecord'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'reason': ('django.db.models.fields.TextField', [], {})
        },
        u'dbio.unknownitem': {
            'Meta': {'object_name': 'UnknownItem', 'db_table': "'wq_unknownitem'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        u'relate.relationship': {
            'Meta': {'object_name': 'Relationship', 'db_table': "'wq_relationship'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            'from_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'to_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            'to_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['relate.RelationshipType']"})
        },
        u'relate.relationshiptype': {
            'Meta': {'object_name': 'RelationshipType', 'db_table': "'wq_relationshiptype'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inverse_name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'to_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"})
        }
    }

    complete_apps = []



from wq.db.patterns import admin
from wq.db.patterns.models import Relationship
from .models import MetaColumn, Range


class MetaColumnAdmin(admin.IdentifiedModelAdmin):
    list_filter = ['type']


class RangeInline(admin.TabularInline):
    model = Range
    extra = 0


class RelationshipAdmin(admin.ModelAdmin):
    inlines = [RangeInline]

admin.site.register(MetaColumn, MetaColumnAdmin)
admin.site.register(Relationship, RelationshipAdmin)



from django.db import models


class NaturalKeyModelManager(models.Manager):
    """
    Manager for use with subclasses of NaturalKeyModel.
    """

    def get_by_natural_key(self, *args):
        """
        Return the object corresponding to the provided natural key.

        (This is a generic implementation of the standard Django function)
        """

        kwargs = self.natural_key_kwargs(*args)

        # Since kwargs already has __ lookups in it, we could just do this:
        # return self.get(**kwargs)

        # But, we should call each related model's get_by_natural_key in case
        # it's been overridden
        for name, rel_to in self.model.get_natural_key_info():
            if not rel_to:
                continue

            # Extract natural key for related object
            nested_key = extract_nested_key(kwargs, rel_to, name)
            if nested_key:
                # Update kwargs with related object
                try:
                    kwargs[name] = rel_to.objects.get_by_natural_key(
                        *nested_key
                    )
                except rel_to.DoesNotExist:
                    # If related object doesn't exist, assume this one doesn't
                    raise self.model.DoesNotExist()
            else:
                kwargs[name] = None

        return self.get(**kwargs)

    def create_by_natural_key(self, *args):
        """
        Create a new object from the provided natural key values.  If the
        natural key contains related objects, recursively get or create them by
        their natural keys.
        """

        kwargs = self.natural_key_kwargs(*args)
        for name, rel_to in self.model.get_natural_key_info():
            if not rel_to:
                continue
            nested_key = extract_nested_key(kwargs, rel_to, name)
            # Automatically create any related objects as needed
            if nested_key:
                kwargs[name], is_new = (
                    rel_to.objects.get_or_create_by_natural_key(*nested_key)
                )
            else:
                kwargs[name] = None
        return self.create(**kwargs)

    def get_or_create_by_natural_key(self, *args):
        """
        get_or_create + get_by_natural_key
        """
        try:
            return self.get_by_natural_key(*args), False
        except self.model.DoesNotExist:
            return self.create_by_natural_key(*args), True

    # Shortcut for common use case
    def find(self, *args):
        """
        Shortcut for get_or_create_by_natural_key that discards the "created"
        boolean.
        """
        obj, is_new = self.get_or_create_by_natural_key(*args)
        return obj

    def natural_key_kwargs(self, *args):
        """
        Convert args into kwargs by merging with model's natural key fieldnames
        """
        natural_key = self.model.get_natural_key_fields()
        if len(args) != len(natural_key):
            raise TypeError("Wrong number of values, expected %s"
                            % len(natural_key))
        return dict(zip(natural_key, args))

    def resolve_keys(self, keys, auto_create=False):
        """
        Resolve the list of given keys into objects, if possible.
        Returns a mapping and a success indicator.
        """
        resolved = {}
        success = True
        for key in keys:
            if auto_create:
                resolved[key] = self.find(*key)
            else:
                try:
                    resolved[key] = self.get_by_natural_key(*key)
                except self.model.DoesNotExist:
                    success = False
                    resolved[key] = None
        return resolved, success


class NaturalKeyModel(models.Model):
    """
    Abstract class with a generic implementation of natural_key.
    """

    objects = NaturalKeyModelManager()

    @classmethod
    def get_natural_key_info(cls):
        """
        Derive natural key from first unique_together definition, noting which
        fields are related objects vs. regular fields.
        """
        fields = cls._meta.unique_together[0]
        info = []
        for name in fields:
            field = cls._meta.get_field_by_name(name)[0]
            rel_to = field.rel.to if field.rel else None
            info.append((name, rel_to))
        return info

    @classmethod
    def get_natural_key_fields(cls):
        """
        Determine actual natural key field list, incorporating the natural keys
        of related objects as needed.
        """
        natural_key = []
        for name, rel_to in cls.get_natural_key_info():
            if not rel_to:
                natural_key.append(name)
            else:
                nested_key = rel_to.get_natural_key_fields()
                natural_key.extend([
                    name + '__' + nname
                    for nname in nested_key
                ])
        return natural_key

    def natural_key(self):
        """
        Return the natural key for this object.

        (This is a generic implementation of the standard Django function)
        """
        # Recursively extract properties from related objects if needed
        vals = [reduce(getattr, name.split('__'), self)
                for name in self.get_natural_key_fields()]
        return vals

    class Meta:
        abstract = True


def extract_nested_key(key, cls, prefix=''):
    nested_key = cls.get_natural_key_fields()
    local_fields = {field.name: field for field in cls._meta.local_fields}
    values = []
    has_val = False
    if prefix:
        prefix += '__'
    for nname in nested_key:
        val = key.pop(prefix + nname, None)
        if val is None and nname.endswith('__primary_identifiers__slug'):
            nname = nname.replace('__primary_identifiers__slug', '')
            val = key.pop(prefix + nname, None)
        if val is None and nname in local_fields:
            if type(local_fields[nname]).__name__ == 'DateTimeField':
                date = key.pop(nname + '_date', None)
                time = key.pop(nname + '_time', None)
                if date and time:
                    val = '%s %s' % (date, time)

        if val is not None:
            has_val = True
        values.append(val)
    if has_val:
        return values
    else:
        return None



from django.contrib.contenttypes.models import (
    ContentType as DjangoContentType,
    ContentTypeManager as DjangoContentTypeManager
)
from wq.db.patterns.models import (
    AnnotatedModel, IdentifiedModel, LocatedModel, RelatedModel
)
from wq.db.patterns.models import RelationshipType
from wq.db.patterns.models import BaseAnnotationType, BaseAnnotation


class ContentTypeManager(DjangoContentTypeManager):
    def get_by_identifier(self, identifier):
        return self.get(model=identifier)


class ContentType(DjangoContentType):
    objects = ContentTypeManager()

    @property
    def identifier(self):
        return self.model

    @property
    def urlbase(self):
        cls = self.model_class()
        if cls is None:
            return None
        config = self.get_config()
        if config:
            return config['url']
        urlbase = unicode(cls._meta.verbose_name_plural)
        return urlbase.replace(' ', '')

    @property
    def is_annotated(self):
        cls = self.model_class()
        return issubclass(cls, AnnotatedModel)

    @property
    def is_annotation(self):
        cls = self.model_class()
        return issubclass(cls, BaseAnnotation)

    @property
    def is_annotationtype(self):
        cls = self.model_class()
        return issubclass(cls, BaseAnnotationType)

    @property
    def is_identified(self):
        cls = self.model_class()
        return issubclass(cls, IdentifiedModel)

    @property
    def is_located(self):
        cls = self.model_class()
        return issubclass(cls, LocatedModel)

    @property
    def is_related(self):
        cls = self.model_class()
        return issubclass(cls, RelatedModel)

    @property
    def has_geo_fields(self):
        cls = self.model_class()
        for fld in ('latitude', 'longitude', 'geometry'):
            if fld in cls._meta.get_all_field_names():
                return True
        return False

    # Get foreign keys for this content type
    def get_foreign_keys(self):
        cls = self.model_class()
        if cls is None:
            return []
        parents = {}
        for f in cls._meta.fields:
            if f.rel is not None and type(f.rel).__name__ == 'ManyToOneRel':
                parent = get_ct(f.rel.to)
                parents.setdefault(parent, [])
                parents[parent].append(f.name)
        return parents

    def get_parents(self):
        return set(self.get_foreign_keys().keys())

    # Get foreign keys and RelationshipType parents for this content type
    def get_all_parents(self):
        parents = self.get_parents()
        if not self.is_related:
            return parents
        for rtype in RelationshipType.objects.filter(to_type=self):
            ctype = rtype.from_type
            # This is a DjangoContentType, swap for our custom version
            ctype = ContentType.objects.get_for_id(ctype.pk)
            parents.add(ctype)
        return parents

    def get_children(self, include_rels=False):
        cls = self.model_class()
        if cls is None:
            return []
        rels = cls._meta.get_all_related_objects()
        if include_rels:
            return set([(get_ct(rel.model), rel) for rel in rels])
        else:
            return set([get_ct(rel.model) for rel in rels])

    def get_all_children(self):
        children = self.get_children()
        if not self.is_related:
            return children
        for rtype in RelationshipType.objects.filter(from_type=self):
            ctype = rtype.to_type
            # This is a DjangoContentType, swap for our custom version
            ctype = ContentType.objects.get_for_id(ctype.pk)
            children.add(ctype)
        return children

    def get_config(self, user=None):
        from .app import router  # avoid circular import
        cls = self.model_class()
        return router.get_model_config(cls, user)

    def is_registered(self):
        from .app import router  # avoid circular import
        cls = self.model_class()
        return router.model_is_registered(cls)

    class Meta:
        proxy = True


def get_ct(model, for_concrete_model=False):
    if isinstance(model, basestring):
        ctype = ContentType.objects.get_by_identifier(model)
    else:
        ctype = ContentType.objects.get_for_model(
            model, for_concrete_model=for_concrete_model
        )
        # get_for_model sometimes returns a DjangoContentType - caching issue?
        if not isinstance(ctype, ContentType):
            ctype = ContentType.objects.get(pk=ctype.pk)
            ContentType.objects._add_to_cache(ContentType.objects.db, ctype)
    return ctype


def get_object_id(instance):
    ct = get_ct(instance)
    config = ct.get_config()
    if config and 'lookup' in config:
        return getattr(instance, config['lookup'])
    elif ct.is_identified and instance.primary_identifier:
        return instance.primary_identifier.slug
    return instance.pk


def get_by_identifier(queryset, ident):
    if hasattr(queryset, 'get_by_identifier'):
        return queryset.get_by_identifier(ident)
    else:
        ct = get_ct(queryset.model)
        config = ct.get_config()
        if config and 'lookup' in config:
            lookup = config['lookup']
        else:
            lookup = 'pk'
        return queryset.get(**{lookup: ident})


class MultiQuerySet(object):
    querysets = []

    def __init__(self, *args, **kwargs):
        self.querysets = args

    def __getitem__(self, index):
        if isinstance(index, slice):
            multi = True
        else:
            multi = False
            index = slice(index, index + 1)

        result = []
        for qs in self.querysets:
            if index.start < qs.count():
                result.extend(qs[index])
            index = slice(index.start - qs.count(),
                          index.stop - qs.count())
            if index.start < 0:
                if index.stop < 0:
                    break
                index = slice(0, index.stop)
        if multi:
            return (item for item in result)
        else:
            return result[0]

    def __iter__(self):
        for qs in self.querysets:
            for item in qs:
                yield item

    def count(self):
        result = 0
        for qs in self.querysets:
            result += qs.count()
        return result

    def __len__(self):
        return self.count()



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from wq.db.patterns.base import SerializableGenericRelation


class RelatedModelManager(models.Manager):
    def filter_by_related(self, *args, **kwargs):
        objects = {}
        for obj in args:
            ctype = ContentType.objects.get_for_model(obj)
            if ctype not in objects:
                objects[ctype] = []
            objects[ctype].append(obj)

        data = self.all()
        for ctype, objs in objects.items():
            if kwargs.get('inverse', False):
                data = data.filter(relationships__to_content_type=ctype,
                                   relationships__to_object_id__in=
                                   [obj.pk for obj in objs])
            else:
                data = data.filter(
                    inverserelationships__from_content_type=ctype,
                    inverserelationships__from_object_id__in=
                    [obj.pk for obj in objs]
                )
        return data


class RelatedModel(models.Model):
    relationships = SerializableGenericRelation(
        'Relationship',
        content_type_field='from_content_type',
        object_id_field='from_object_id',
        related_name='%(class)s_set'
    )
    inverserelationships = SerializableGenericRelation(
        'InverseRelationship',
        content_type_field='to_content_type',
        object_id_field='to_object_id',
        related_name='%(class)s_inverse_set'
    )
    objects = RelatedModelManager()

    def all_relationships(self):
        for rel in self.relationships.all():
            yield rel
        for rel in self.inverserelationships.all():
            yield rel

    def create_relationship(self, to_obj, name,
                            inverse_name=None, computed=False):
        return Relationship.objects.create_relationship(
            from_obj=self,
            to_obj=to_obj,
            name=name,
            inverse_name=inverse_name,
            computed=computed
        )

    class Meta:
        abstract = True


class RelationshipManager(models.Manager):
    def create_relationship(self, from_obj, to_obj, name,
                            inverse_name=None, computed=False):
        from_ct = ContentType.objects.get_for_model(from_obj)
        to_ct = ContentType.objects.get_for_model(to_obj)
        reltype, is_new = RelationshipType.objects.get_or_create(
            from_type=from_ct,
            to_type=to_ct,
            name=name,
            inverse_name=inverse_name,
            computed=computed,
        )
        rel, is_new = self.get_or_create(
            type=reltype,
            from_content_type=from_ct,
            from_object_id=from_obj.pk,
            to_content_type=to_ct,
            to_object_id=to_obj.pk,
            computed=computed,
        )
        return rel


class Relationship(models.Model):

    type = models.ForeignKey('RelationshipType')

    # Objects can contain pointers to any model
    from_content_type = models.ForeignKey(ContentType, related_name='+')
    from_object_id = models.PositiveIntegerField()
    from_content_object = generic.GenericForeignKey(
        'from_content_type', 'from_object_id'
    )

    to_content_type = models.ForeignKey(ContentType, related_name='+')
    to_object_id = models.PositiveIntegerField()
    to_content_object = generic.GenericForeignKey(
        'to_content_type', 'to_object_id'
    )

    computed = models.BooleanField(default=False)

    objects = RelationshipManager()

    _dict_cache = {}
    _cache_prefix = 'rel'

    @property
    def left(self):
        return self.from_content_object

    @property
    def right(self):
        return self.to_content_object

    @right.setter
    def right(self, value):
        self.to_content_object = value

    @property
    def right_object_id(self):
        return self.to_object_id

    @property
    def right_content_type_id(self):
        return self.to_content_type_id

    @property
    def reltype(self):
        return RelationshipType.objects.get_type(self.type_id)

    @property
    def right_dict(self):
        key = self._cache_prefix + str(self.pk)
        cache = type(self)._dict_cache
        if key not in cache:
            from wq.db.rest.models import get_object_id, get_ct
            obj = self.right
            oid = get_object_id(obj)
            ct = get_ct(obj)
            cache[key] = {
                'item_id': get_object_id(obj),
                'item_label': unicode(obj),
                'item_url': '%s/%s' % (ct.urlbase, oid),
                'item_page': ct.identifier
            }
        return cache[key]

    def save(self, *args, **kwargs):
        rightct = self.reltype.right
        self.right = rightct.get_object_for_this_type(pk=self.right_object_id)
        super(Relationship, self).save(*args, **kwargs)

    def __unicode__(self):
        if (self.from_content_type_id and self.type_id
                and self.to_content_type_id):
            return u'%s %s %s' % (self.left, self.reltype, self.right)
        else:
            return 'Undefined'

    class Meta:
        db_table = 'wq_relationship'


class InverseRelationshipManager(models.Manager):
    def get_query_set(self):
        qs = super(InverseRelationshipManager, self).get_query_set()
        return qs.filter(type__inverse_name__isnull=False)


class InverseRelationship(Relationship):
    objects = InverseRelationshipManager()
    _cache_prefix = 'inv'

    @property
    def left(self):
        return self.to_content_object

    @property
    def right(self):
        return self.from_content_object

    @right.setter
    def right(self, value):
        self.from_content_object = value

    @property
    def right_object_id(self):
        return self.from_object_id

    @property
    def right_content_type_id(self):
        return self.from_content_type_id

    @property
    def reltype(self):
        return InverseRelationshipType.objects.get_type(self.type_id)

    class Meta:
        proxy = True


class RelationshipTypeManager(models.Manager):
    _type_cache = {}
    _cache_prefix = 'rel'

    def get_type(self, pk):
        key = self._cache_prefix + str(pk)
        cache = type(self)._type_cache
        if key not in cache:
            cache[key] = self.get(pk=pk)
        return cache[key]


class RelationshipType(models.Model):
    name = models.CharField(max_length=255)
    inverse_name = models.CharField(max_length=255, null=True, blank=True)

    from_type = models.ForeignKey(ContentType, related_name='+')
    to_type = models.ForeignKey(ContentType, related_name='+')

    computed = models.BooleanField(default=False)

    objects = RelationshipTypeManager()

    @property
    def left(self):
        return self.from_type

    @property
    def right(self):
        return self.to_type

    def __unicode__(self):
        return self.name

    class Meta:
        db_table = 'wq_relationshiptype'


class InverseRelationshipTypeManager(RelationshipTypeManager):
    _cache_prefix = 'inv'

    def get_query_set(self):
        qs = super(InverseRelationshipTypeManager, self).get_query_set()
        return qs.filter(inverse_name__isnull=False)


class InverseRelationshipType(RelationshipType):
    objects = InverseRelationshipTypeManager()

    @property
    def left(self):
        return self.to_type

    @property
    def right(self):
        return self.from_type

    def __unicode__(self):
        return self.inverse_name

    class Meta:
        proxy = True



from django.contrib.contenttypes import generic
from django.contrib import admin
from django import forms

from .models import (
    Relationship, InverseRelationship, RelatedModel,
    RelationshipType, InverseRelationshipType
)


class RelationshipForm(forms.ModelForm):
    type = forms.models.ModelChoiceField(
        queryset=RelationshipType.objects.filter(computed=False)
    )


class InverseRelationshipForm(forms.ModelForm):
    type = forms.models.ModelChoiceField(
        queryset=InverseRelationshipType.objects.filter(computed=False)
    )


class RelationshipInline(generic.GenericTabularInline):
    model = Relationship
    ct_field = "from_content_type"
    ct_fk_field = "from_object_id"
    form = RelationshipForm
    extra = 0
    readonly_fields = ('to_content_object',)

    def queryset(self, request):
        return Relationship.objects.filter(computed=False)


class InverseRelationshipInline(generic.GenericTabularInline):
    model = InverseRelationship
    ct_field = "to_content_type"
    ct_fk_field = "to_object_id"
    form = InverseRelationshipForm
    extra = 0
    readonly_fields = ('from_content_object',)

    def queryset(self, request):
        return InverseRelationship.objects.filter(computed=False)


class RelatedModelAdmin(admin.ModelAdmin):
    model = RelatedModel
    inlines = [
        RelationshipInline,
        InverseRelationshipInline
    ]



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models
import swapper


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Adding model 'FileType'
        db.create_table('wq_filetype', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('mimetype', self.gf('django.db.models.fields.CharField')(max_length=255)),
        ))
        db.send_create_signal('files', ['FileType'])

        if not swapper.is_swapped('files', 'File'):
            # Adding model 'File'
            User = swapper.load_model('auth', 'User')
            db.create_table('wq_file', (
                ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('name', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
                ('type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['files.FileType'], null=True, blank=True)),
                ('file', self.gf('wq.db.contrib.files.models.FileField')(max_length=100)),
                ('size', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
                ('width', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
                ('height', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
                ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=User, null=True, blank=True)),
            ))
            db.send_create_signal('files', ['File'])

    def backwards(self, orm):

        # Deleting model 'FileType'
        db.delete_table('wq_filetype')

        if not swapper.is_swapped('files', 'File'):
            # Deleting model 'File'
            db.delete_table('wq_file')

    models = {
        u'annotate.annotation': {
            'Meta': {'object_name': 'Annotation', 'db_table': "'wq_annotation'"},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['annotate.AnnotationType']"}),
            'value': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        u'annotate.annotationtype': {
            'Meta': {'object_name': 'AnnotationType', 'db_table': "'wq_annotationtype'"},
            'contenttype': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']", 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'files.file': {
            'Meta': {'object_name': 'File', 'db_table': "'wq_file'"},
            'file': ('wq.db.contrib.files.models.FileField', [], {'max_length': '100'}),
            'height': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'size': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['files.FileType']", 'null': 'True', 'blank': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']", 'null': 'True', 'blank': 'True'}),
            'width': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'})
        },
        u'files.filetype': {
            'Meta': {'object_name': 'FileType', 'db_table': "'wq_filetype'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'mimetype': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        u'relate.relationship': {
            'Meta': {'object_name': 'Relationship', 'db_table': "'wq_relationship'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            'from_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'to_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            'to_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['relate.RelationshipType']"})
        },
        u'relate.relationshiptype': {
            'Meta': {'object_name': 'RelationshipType', 'db_table': "'wq_relationshiptype'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inverse_name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'to_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"})
        }
    }

    complete_apps = ['files']



from django.contrib.contenttypes import generic
from django.contrib import admin

from .models import Identifier, IdentifiedModel, Authority


class IdentifierInline(generic.GenericTabularInline):
    model = Identifier
    prepopulated_fields = {"slug": ("name",)}
    extra = 0


class IdentifiedModelAdmin(admin.ModelAdmin):
    model = IdentifiedModel
    inlines = [
        IdentifierInline
    ]


class AuthorityIdentifierInline(admin.TabularInline):
    model = Identifier
    extra = 0


class AuthorityAdmin(admin.ModelAdmin):
    model = Authority
    inlines = [
        AuthorityIdentifierInline
    ]



from rest_framework.response import Response
from rest_framework.decorators import link, action
from wq.db.rest.views import ModelViewSet
from wq.db.contrib.dbio import tasks
from wq.io.exceptions import *
from celery.result import AsyncResult


class IoViewSet(ModelViewSet):
    @link()
    def status(self, request, *args, **kwargs):
        taskid = request.GET.get('task', None)
        if not taskid:
            return Response({})

        result = AsyncResult(taskid)
        response = {
            'status': result.state
        }
        if result.state in ('PROGRESS', 'SUCCESS'):
            response.update(result.result)
        elif result.state == 'FAILURE':
            response['error'] = str(result.result)
        return Response(response)

    def run_task(self, name, async=False):
        response = self.retrieve(self.request, **self.kwargs)
        task = getattr(tasks, name).delay(
            self.get_instance(), self.request.user
        )
        if async:
            response.data['task_id'] = task.task_id
        else:
            try:
                response.data['result'] = task.get()
            except IoException as e:
                response.data['error'] = str(e)

        return response

    @link()
    def start(self, request, *args, **kwargs):
        self.action = 'columns'
        return self.run_task('read_columns')

    @action()
    def columns(self, request, *args, **kwargs):
        response = self.run_task('read_columns')
        result = tasks.update_columns.delay(
            self.get_instance(), request.user, request.POST
        )
        response.data['result'] = result.get()
        return response

    @link()
    def ids(self, request, *args, **kwargs):
        response = self.run_task('read_row_identifiers')
        return response

    @action()
    def updateids(self, request, *args, **kwargs):
        response = self.run_task('read_row_identifiers')
        self.action = 'ids'
        result = tasks.update_row_identifiers.delay(
            self.get_instance(), request.user, request.POST
        )
        response.data['result'] = result.get()
        return response

    @action()
    def reset(self, request, *args, **kwargs):
        self.task = 'retrieve'
        response = self.run_task('reset')
        return response

    @action()
    def data(self, request, *args, **kwargs):
        return self.run_task('import_data', async=True)

    @action()
    def auto(self, request, *args, **kwargs):
        return self.run_task('auto_import', async=True)

    def get_instance(self):
        return self.object



from django.contrib import admin
from .models import File, FileType
from wq.db.patterns.admin import (
    AnnotationInline, RelationshipInline, InverseRelationshipInline
)


class FileAdmin(admin.ModelAdmin):
    inlines = [AnnotationInline, RelationshipInline, InverseRelationshipInline]
    list_display = ('__unicode__', 'type')
    list_filter = ('type', )

admin.site.register(File, FileAdmin)
admin.site.register(FileType)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Adding model 'Relationship'
        db.create_table('wq_relationship', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['relate.RelationshipType'])),
            ('from_content_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', to=orm['contenttypes.ContentType'])),
            ('from_object_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ('to_content_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', to=orm['contenttypes.ContentType'])),
            ('to_object_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ('computed', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal('relate', ['Relationship'])

        # Adding model 'RelationshipType'
        db.create_table('wq_relationshiptype', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('inverse_name', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('from_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', to=orm['contenttypes.ContentType'])),
            ('to_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', to=orm['contenttypes.ContentType'])),
            ('computed', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal('relate', ['RelationshipType'])

        # Create indexes
        db.execute("""CREATE INDEX wq_relationship_from_idx ON wq_relationship
                         (from_content_type_id, from_object_id)""")
        db.execute("""CREATE INDEX wq_relationship_to_idx ON wq_relationship
                         (to_content_type_id, to_object_id)""")

    def backwards(self, orm):

        # Drop indexes
        db.execute("DROP INDEX wq_relationship_to_idx")
        db.execute("DROP INDEX wq_relationship_from_idx")

        # Deleting model 'Relationship'
        db.delete_table('wq_relationship')

        # Deleting model 'RelationshipType'
        db.delete_table('wq_relationshiptype')

    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'relate.relationship': {
            'Meta': {'object_name': 'Relationship', 'db_table': "'wq_relationship'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': "orm['contenttypes.ContentType']"}),
            'from_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'to_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': "orm['contenttypes.ContentType']"}),
            'to_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['relate.RelationshipType']"})
        },
        'relate.relationshiptype': {
            'Meta': {'object_name': 'RelationshipType', 'db_table': "'wq_relationshiptype'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inverse_name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'to_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': "orm['contenttypes.ContentType']"})
        }
    }

    complete_apps = ['relate']



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Adding model 'Identifier'
        db.create_table('wq_identifier', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255, db_index=True)),
            ('slug', self.gf('django.db.models.fields.SlugField')(max_length=50, db_index=True)),
            ('authority', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['identify.Authority'], null=True, blank=True)),
            ('is_primary', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'])),
            ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
        ))
        db.send_create_signal('identify', ['Identifier'])

        # Adding model 'Authority'
        db.create_table('wq_identifiertype', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('homepage', self.gf('django.db.models.fields.URLField')(max_length=200, null=True, blank=True)),
            ('object_url', self.gf('django.db.models.fields.URLField')(max_length=200, null=True, blank=True)),
        ))
        db.send_create_signal('identify', ['Authority'])

        # Create indexes
        db.execute("""CREATE INDEX wq_identifier_idx ON wq_identifier
                         (content_type_id, object_id)""")

    def backwards(self, orm):

        # Drop indexes
        db.execute("DROP INDEX wq_identifier_idx")

        # Deleting model 'Identifier'
        db.delete_table('wq_identifier')

        # Deleting model 'Authority'
        db.delete_table('wq_identifiertype')

    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'identify.authority': {
            'Meta': {'object_name': 'Authority', 'db_table': "'wq_identifiertype'"},
            'homepage': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'object_url': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        'identify.identifier': {
            'Meta': {'object_name': 'Identifier', 'db_table': "'wq_identifier'"},
            'authority': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['identify.Authority']", 'null': 'True', 'blank': 'True'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_primary': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'db_index': 'True'})
        }
    }

    complete_apps = ['identify']



from rest_framework.serializers import ModelSerializer as RestModelSerializer
from rest_framework.serializers import (
    Field, DateTimeField, WritableField, RelatedField, PrimaryKeyRelatedField
)
from rest_framework.pagination import (
    PaginationSerializer as RestPaginationSerializer
)

from django.contrib.gis.db.models import fields
from django.contrib.gis.geos import GEOSGeometry
from django.utils import timezone

from django.conf import settings

from .models import (
    DjangoContentType, ContentType,
    get_ct, get_object_id, get_by_identifier
)


class GeometryField(WritableField):
    def to_native(self, value):
        if value is None:
            return None
        import json
        return json.loads(value.geojson)

    def from_native(self, value):
        import json
        if isinstance(value, dict):
            value = json.dumps(value)
        geom = GEOSGeometry(value)
        srid = getattr(settings, 'SRID', 4326)

        if 'crs' in value and value['crs'].get('type', None) == 'name':
            name = value['crs']['properties']['name']
            if name.startswith('urn:ogc:def:crs:EPSG::'):
                geom.srid = int(name.replace('urn:ogc:def:crs:EPSG::', ''))

        if geom.srid is None:
            geom.srid = 4326
        if geom.srid != srid:
            geom.transform(srid)
        return geom


class LabelRelatedField(RelatedField):
    def field_to_native(self, obj, field_name):
        if field_name.endswith('_label'):
            field_name = field_name[:-6]
            if getattr(obj, field_name + '_id') is None:
                return None
        val = getattr(obj, self.source or field_name)
        if self.many:
            return [unicode(item) for item in val.all()]
        else:
            return unicode(val)


class IDField(Field):
    def field_to_native(self, obj, field_name):
        return get_object_id(obj)


#TODO: investigate use of SlugRelatedField and/or PrimaryKeyRelatedField
class IDRelatedField(RelatedField):
    read_only = False

    def field_to_native(self, obj, field_name):
        if field_name.endswith('_id'):
            if getattr(obj, field_name) is None:
                return None
            field_name = field_name[:-3]
        val = getattr(obj, self.source or field_name)
        if val is None:
            return None
        if self.many:
            return [get_object_id(item) for item in val.all()]
        else:
            return get_object_id(val)

    def field_from_native(self, data, files, field_name, into):
        if field_name.endswith('_id'):
            field_name = field_name[:-3]
        return super(IDRelatedField, self).field_from_native(
            data, files, field_name, into
        )

    def from_native(self, data):
        return get_by_identifier(self.queryset, data)


class ContentTypeField(RelatedField):
    read_only = False

    def to_native(self, obj):
        return ContentType.objects.get(pk=obj.pk).identifier

    def from_native(self, data, files):
        return get_ct(data)


class LocalDateTimeField(Field):
    read_only = True

    def to_native(self, obj):
        if obj is None:
            return None
        if obj.tzinfo:
            obj = timezone.localtime(obj)
        return obj.strftime('%Y-%m-%d %I:%M %p')


class ModelSerializer(RestModelSerializer):
    def __init__(self, *args, **kwargs):
        for field in ('Geometry', 'GeometryCollection',
                      'Point', 'LineString', 'Polygon',
                      'MultiPoint', 'MultiLineString', 'MultiPolygon'):
            self.field_mapping[
                getattr(fields, field + 'Field')
            ] = GeometryField
        super(ModelSerializer, self).__init__(*args, **kwargs)

    @property
    def router(self):
        if 'view' not in self.context and 'router' not in self.context:
            return None
        if 'view' in self.context:
            return self.context['view'].router
        else:
            return self.context['router']

    def get_default_fields(self, *args, **kwargs):
        fields = super(ModelSerializer, self).get_default_fields(
            *args, **kwargs
        )
        fields['id'] = IDField()
        if 'label' not in self.opts.exclude:
            fields['label'] = Field(source='__unicode__')

        if not self.router:
            return fields

        nested = self.context.get('nested', False)
        many = self.many or self.source == 'object_list'
        if 'request' in self.context:
            renderer = self.context['request'].accepted_renderer
            geo = renderer.format == 'geojson'
            saving = self.context['request'].method != 'GET'
        else:
            geo = False
            saving = False
        saving = self.context.get('saving', saving)

        def add_labels(name, field, qs):
            # Add _id and _label for context
            fields[name + '_id'] = IDRelatedField(
                source=name,
                required=field.required,
                queryset=qs
            )
            fields[name + '_label'] = LabelRelatedField(queryset=qs)

        # Special handling for related fields
        for name, field in fields.items():
            if name == 'label':
                continue

            model_field, model, direct, m2m = (
                self.opts.model._meta.get_field_by_name(name)
            )

            if isinstance(field, DateTimeField):
                fields[name + '_label'] = LocalDateTimeField(name)

            if model_field.choices:
                fields[name + '_label'] = Field('get_%s_display' % name)

            if (not isinstance(field, ModelSerializer)
                    and not isinstance(field, PrimaryKeyRelatedField)):
                continue

            if model_field.rel.to == DjangoContentType:
                fields['for'] = ContentTypeField(
                    source=name,
                    queryset=getattr(
                        field, 'object', getattr(field, 'queryset', None)
                    )
                )
                del fields[name]
                continue

            if (self.opts.depth < 1 and not (saving and m2m)
                    or (saving and not m2m) or nested):
                # In list views, remove [fieldname] as an attribute in favor of
                # [fieldname]_id and [fieldname]_label (below).
                # (Except when saving m2m items, then we need the nested field)
                del fields[name]
            else:
                # In detail views, always make [fieldname] a nested field. This
                # is needed to generate renderer contexts to support deeplinks.
                # (The list view doesn't need the full object, as REST clients
                #  like wq.app's app.js can use preloaded lists and callbacks
                #  to resolve [fieldname] to an object, given [fieldname]_id.)
                fields[name] = self.get_nested_field(model_field)

            # Stop processing if this is a many-to-many-field
            if field.many or m2m:
                continue
            else:
                if isinstance(field, PrimaryKeyRelatedField):
                    add_labels(name, field, field.queryset)
                else:
                    add_labels(name, field, field.opts.model.objects)

        # Add child objects (serialize with registered serializers)
        if (not self.opts.depth and not saving and not geo) or nested:
            return fields

        for vf in self.opts.model._meta.virtual_fields:
            if not self.opts.depth and geo and vf.name != "locations":
                continue
            if getattr(vf, 'serialize', False) and vf.name not in fields:
                cls = self.router.get_serializer_for_model(
                    vf.rel.to, max(self.opts.depth - 1, 0)
                )
                context = self.context.copy()
                context['nested'] = True
                fields[vf.name] = cls(context=context, many=True)

        if geo and not self.opts.depth:
            return fields

        ct = get_ct(self.opts.model)
        for cct, rel in ct.get_children(include_rels=True):
            accessor = rel.get_accessor_name()
            if accessor == rel.field.related_query_name():
                cls = self.router.get_serializer_for_model(
                    cct.model_class(), max(self.opts.depth - 1, 0)
                )
                context = self.context.copy()
                context['nested'] = True
                many = not rel.field.unique  # Check for OneToOneField
                fields[accessor] = cls(context=context, many=many)

        return fields

    def get_nested_field(self, model_field):
        model = model_field.rel.to
        if self.router:
            cls = self.router.get_serializer_for_model(
                model, self.opts.depth - 1
            )
            return cls(
                context=self.context,
                required=not(model_field.null or model_field.blank)
            )
        return super(ModelSerializer, self).get_nested_field(model_field)

    # Any IDRelatedField errors should be reported without the '_id' suffix
    # FIXME: is there a better way to fix this?
    @property
    def errors(self):
        errors = super(ModelSerializer, self).errors
        for field_name in self.fields:
            if field_name not in errors or not field_name.endswith('_id'):
                continue
            if not type(self.fields[field_name]) == IDRelatedField:
                continue
            field_name = field_name[:-3]
            errors[field_name] = errors[field_name + '_id']
            del errors[field_name + '_id']
        return errors


class PaginationSerializer(RestPaginationSerializer):
    results_field = 'list'
    page = Field('number')
    pages = Field('paginator.num_pages')
    per_page = Field('paginator.per_page')

    def to_native(self, obj):
        result = super(PaginationSerializer, self).to_native(obj)
        result['multiple'] = result['pages'] > 1
        return result



from rest_framework import serializers
from rest_pandas import PandasSerializer
from wq.db.rest.serializers import LocalDateTimeField
from django.db.models.fields import DateTimeField

import swapper
EventResult = swapper.load_model('vera', 'EventResult')
Event = swapper.load_model('vera', 'Event')
EVENT_INDEX = Event._meta.unique_together[0]


class EventResultSerializer(PandasSerializer):
    index_none_value = "-"

    parameter = serializers.Field(source="result_type.primary_identifier.slug")
    units = serializers.Field(source="result_type.units")
    value = serializers.Field(source="result_value")

    def get_default_fields(self):
        """"
        Map fields in EVENT_INDEX to actual natural key lookup values.
        E.g. the default Event has two natural key fields, assuming
        Site is an IdentifiedModel:
            site -> event_site.primary_identifier.slug
            date -> event_date
        """
        fields = {}
        lookups = Event.get_natural_key_fields()
        for key, lookup in zip(EVENT_INDEX, lookups):
            field = Event._meta.get_field_by_name(key)[0]
            lookup = lookup.replace('__', '.')
            lookup = lookup.replace(
                "primary_identifiers.", "primary_identifier."
            )
            if isinstance(field, DateTimeField):
                fields[key] = LocalDateTimeField("event_" + lookup)
            else:
                fields[key] = serializers.Field("event_" + lookup)
        return fields

    def get_index(self, dataframe):
        """
        By default, all fields from Event's natural key need to be included or
        pivoting may break due to non-unique values.  Move first item in index
        (which would usually be "site") to the end to facilitate unstacking.
        """
        index_fields = []
        for key in EVENT_INDEX:
            if key not in self.opts.exclude:
                index_fields.append(key)

        return index_fields[1:] + [index_fields[0], 'parameter', 'units']

    def get_dataframe(self, data):
        """
        Unstack the dataframe so units, parameter, and site are columns.
        """
        dataframe = super(EventResultSerializer, self).get_dataframe(data)
        dataframe.columns.name = ""
        dataframe = (
            dataframe
            .unstack().unstack().unstack()
            .dropna(axis=0, how='all')
            .dropna(axis=1, how='all')
        )
        return dataframe

    class Meta:
        model = EventResult



from django.contrib.contenttypes.generic import GenericRelation
from south.modelsinspector import add_ignored_fields


# Trick rest_framework into serializing these relationships
class SerializableGenericRelation(GenericRelation):
    def __init__(self, *args, **kwargs):
        super(SerializableGenericRelation, self).__init__(*args, **kwargs)
        self.serialize = True

add_ignored_fields(["^wq.db.patterns.base.SerializableGenericRelation"])



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

from wq.db.patterns.locate.models import SRID


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Adding model 'Location'
        db.create_table('wq_location', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('is_primary', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('geometry', self.gf('django.contrib.gis.db.models.fields.GeometryField')(srid=SRID)),
            ('accuracy', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'])),
            ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
        ))
        db.send_create_signal('locate', ['Location'])

        # Create index
        db.execute("""CREATE INDEX wq_location_idx ON wq_location
                         (content_type_id, object_id)""")

    def backwards(self, orm):

        # Drop index
        db.execute("DROP INDEX wq_location_idx;")

        # Deleting model 'Location'
        db.delete_table('wq_location')

    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'locate.location': {
            'Meta': {'object_name': 'Location', 'db_table': "'wq_location'"},
            'accuracy': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'geometry': ('django.contrib.gis.db.models.fields.GeometryField', [], {'srid': SRID}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_primary': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {})
        }
    }

    complete_apps = ['locate']



from wq.db.rest import app
from .models import Identifier, Authority
from .serializers import IdentifierSerializer

app.router.register_model(Identifier, serializer=IdentifierSerializer)
app.router.register_model(Authority)



from wq.db.rest import app
from wq.db.rest.serializers import ModelSerializer
from django.contrib.auth.models import User
from django.conf import settings

HAS_SOCIAL_AUTH = ('social.apps.django_app.default' in settings.INSTALLED_APPS)


class UserSerializer(ModelSerializer):
    def to_native(self, instance):
        result = super(UserSerializer, self).to_native(instance)
        if (HAS_SOCIAL_AUTH and 'social_auth' in result):
            if len(result['social_auth']) > 0:
                result['social_auth'] = {'accounts': result['social_auth']}
            else:
                result['social_auth'] = None
        return result

    class Meta:
        exclude = ('id', 'password')


class SocialAuthSerializer(ModelSerializer):
    def to_native(self, instance):
        return {
            'provider_id': unicode(instance.provider),
            'provider_label': instance.provider.title(),
            'id': instance.pk,
            'label': instance.uid,
            'can_disconnect': type(instance).allowed_to_disconnect(
                user=instance.user,
                backend_name=instance.provider,
                association_id=instance.pk
            )
        }

app.router.register_serializer(User, UserSerializer)
if HAS_SOCIAL_AUTH:
    from social.apps.django_app.default.models import UserSocialAuth
    app.router.register_serializer(UserSocialAuth, SocialAuthSerializer)



from django.contrib.contenttypes import generic
from django.contrib.gis import admin
from django.contrib.gis.db import models
from .models import LocatedModel, Location, SRID


class LocationInline(generic.GenericTabularInline):
    model = Location
    extra = 1  # Otherwise OpenLayers will not initialize properly...

    # Mimics GeoModelAdmin.formfield_for_dbfield
    # (since there is not an Inline version of GeoModelAdmin)
    def formfield_for_dbfield(self, db_field, **kwargs):
        if isinstance(db_field, models.GeometryField):
            request = kwargs.pop('request', None)
            # Use parent (LocatedModelAdmin) class to actually get_map_widget
            kwargs['widget'] = self.admin_site._registry.get(
                self.parent_model
            ).get_map_widget(
                db_field
            )
            return db_field.formfield(**kwargs)
        else:
            return super(LocationInline, self).formfield_for_dbfield(
                db_field, **kwargs
            )


class LocatedModelAdmin(admin.GeoModelAdmin):
    inlines = [LocationInline]
    geom_type = 'POINT'
    map_width = 300
    map_height = 200
    map_srid = SRID
    default_zoom = 1
    point_zoom = 10

    # GeoModelAdmin does not know how to handle generic GEOMETRY types;
    # Need to explicitly set geom_type
    def get_map_widget(self, db_field):
        "Overrides db_field geometry type to ensure a meaningful editor"
        class FakeDbField(object):
            geom_type = self.geom_type
            name = db_field.name
        return super(LocatedModelAdmin, self).get_map_widget(FakeDbField())



from wq.db.rest.serializers import ModelSerializer
from wq.db.rest.models import get_ct, get_object_id
import json
from .models import Location


class LocationSerializer(ModelSerializer):

    @property
    def as_geometry(self):
        if 'request' in self.context:
            renderer = self.context['request'].accepted_renderer
            if renderer.format == 'geojson':
                return True
        return False

    def to_native(self, loc):
        has_parent = self.parent and hasattr(self.parent.opts, 'model')
        if self.as_geometry and has_parent:
            return json.loads(loc.geometry.geojson)

        data = super(LocationSerializer, self).to_native(loc)
        if has_parent:
            pass
        else:
            # Include pointer to parent object (see annotate/serializers.py)
            idname = get_ct(loc.content_object).identifier + '_id'
            data[idname] = get_object_id(loc.content_object)
        return data

    def from_native(self, data, files):
        if (data.get('type', None) == 'Feature'
                and 'properties' in data and 'geometry' in data):
            obj = data['properties']
            obj['geometry'] = data['geometry']
            if 'id' in data:
                obj['id'] = data['id']
            if 'crs' in data:
                obj['geometry']['crs'] = data['crs']
        return super(LocationSerializer, self).from_native(data, files)

    def field_to_native(self, obj, field_name):
        if not self.as_geometry:
            return super(LocationSerializer, self).field_to_native(
                obj, field_name
            )

        loc = getattr(obj, field_name or self.source)
        if loc.count() == 1:
            return self.to_native(loc.all()[0])
        elif loc.count() > 1:
            return {
                'type': 'GeometryCollection',
                'geometries': [
                    self.to_native(l) for l in loc.all()
                ]
            }
        else:
            return None

    def field_from_native(self, data, files, field_name, into):
        vals = data.get(field_name, None)
        if isinstance(vals, basestring) and vals.strip() != '':
            vals = json.loads(vals)
        if (isinstance(vals, dict) and
                vals.get('type', None) == "FeatureCollection"):
            if 'crs' in vals:
                features = []
                for feature in vals['features']:
                    feature.update({
                        'crs': vals['crs']
                    })
                    features.append(feature)
            else:
                features = vals['features']
            data = {field_name: features}
        else:
            data = {field_name: []}
        return super(LocationSerializer, self).field_from_native(
            data, files, field_name, into
        )

    class Meta:
        exclude = ('content_type_id', 'for', 'object_id')



from wq.db.rest import app
from .models import Relationship, InverseRelationship, RelationshipType
from .serializers import (
    RelationshipSerializer,
    InverseRelationshipSerializer,
    RelationshipTypeSerializer
)

app.router.register_model(
    Relationship, serializer=RelationshipSerializer
)
app.router.register_model(
    InverseRelationship, serializer=InverseRelationshipSerializer
)
app.router.register_model(
    RelationshipType, serializer=RelationshipTypeSerializer
)



# Django settings
TEMPLATE_LOADERS = (
    'wq.db.rest.template.Loader',
    'django.template.loaders.app_directories.Loader',
)
TEMPLATE_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
    'wq.db.rest.auth.context_processors.is_authenticated',
    'wq.db.rest.auth.context_processors.social_auth',
    'wq.db.rest.auth.context_processors.csrftoken',
    'wq.db.rest.context_processors.version',
)
SESSION_COOKIE_HTTPONLY = False

# Django Rest Framework settings
REST_FRAMEWORK = {
    'PAGINATE_BY': 50,
    'PAGINATE_BY_PARAM': 'limit',

    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.TemplateHTMLRenderer',
        'wq.db.rest.renderers.JSONRenderer',
        'wq.db.rest.renderers.AMDRenderer',
        'wq.db.rest.renderers.GeoJSONRenderer',
    ),

    'DEFAULT_MODEL_SERIALIZER_CLASS':
    'wq.db.rest.serializers.ModelSerializer',

    'DEFAULT_PAGINATION_SERIALIZER_CLASS':
    'wq.db.rest.serializers.PaginationSerializer',

    'DEFAULT_PERMISSION_CLASSES': (
        'wq.db.rest.permissions.ModelPermissions',
    ),

    'FILTER_BACKEND': 'wq.db.rest.filters.FilterBackend'
}

# Django Social Auth settings
SOCIAL_AUTH_PIPELINE = (
    'social.pipeline.social_auth.social_details',
    'social.pipeline.social_auth.social_uid',
    'social.pipeline.social_auth.auth_allowed',
    'social.pipeline.social_auth.social_user',
    'social.pipeline.user.get_username',
    'social.pipeline.user.create_user',
    'social.pipeline.social_auth.associate_user',
    'social.pipeline.social_auth.load_extra_data',
    'social.pipeline.user.user_details',
    'wq.db.rest.auth.pipeline.assign_default_group',
)

# wq.db settings
ANONYMOUS_PERMISSIONS = tuple()
SRID = 4326
DEFAULT_AUTH_GROUP = "Users"
DISAMBIGUATE = True



from wq.db.patterns import admin
from wq.db.patterns.models import Authority, RelationshipType
import swapper
from django.conf import settings


class ResultInline(admin.TabularInline):
    model = swapper.load_model('vera', 'Result')
    extra = 0


class ReportInline(admin.TabularInline):
    model = swapper.load_model('vera', 'Report')
    extra = 0


class ParameterAdmin(admin.IdentifiedModelAdmin):
    list_display = ('name', 'units')
    list_filter = ('units', 'is_numeric')


class ReportAdmin(admin.ModelAdmin):
    inlines = [ResultInline]


class EventAdmin(admin.ModelAdmin):
    inlines = [ReportInline]

# Register models with admin, but only if they haven't been swapped
default_admin = {
    'Site': admin.IdentifiedRelatedModelAdmin,
    'Event': EventAdmin,
    'Report': ReportAdmin,
    'ReportStatus': admin.ModelAdmin,
    'Parameter': ParameterAdmin,
}
for model in default_admin:
    if swapper.is_swapped('vera', model):
        continue
    admin.site.register(
        swapper.load_model('vera', model),
        default_admin[model]
    )

# Auto-register relevant patterns' type models
admin.site.register(Authority, admin.AuthorityAdmin)
admin.site.register(RelationshipType)
